# -----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------
import io
import os
import re
import struct
import pathlib

from PyInstaller import log as logging
from PyInstaller.archive.writers import SplashWriter
from PyInstaller.building import splash_templates
from PyInstaller.building.datastruct import Target
from PyInstaller.building.utils import _check_guts_eq, _check_guts_toc, misc
from PyInstaller.compat import is_aix, is_darwin
from PyInstaller.depend import bindepend

logger = logging.getLogger(__name__)


class Splash(Target):
    """
    Bundles the required resources for the splash screen into a file, which will be included in the CArchive.

    A Splash has two outputs, one is itself and one is stored in splash.binaries. Both need to be passed to other
    build targets in order to enable the splash screen.
    """
    def __init__(self, image_file, binaries, datas, **kwargs):
        """
        :param str image_file:
            A path-like object to the image to be used. Only the PNG file format is supported.

            .. note:: If a different file format is supplied and PIL (Pillow) is installed, the file will be converted
                automatically.

            .. note:: *Windows*: The color ``'magenta'`` / ``'#ff00ff'`` must not be used in the image or text, as it is
                used by splash screen to indicate transparent areas. Use a similar color (e.g., ``'#ff00fe'``) instead.

            .. note:: If PIL (Pillow) is installed and the image is bigger than max_img_size, the image will be resized
                to fit into the specified area.
        :param list binaries:
            The TOC list of binaries the Analysis build target found. This TOC includes all extension modules and their
            binary dependencies. This is required to determine whether the user's program uses `tkinter`.
        :param list datas:
            The TOC list of data the Analysis build target found. This TOC includes all data-file dependencies of the
            modules. This is required to check if all splash screen requirements can be bundled.

        :keyword text_pos:
            An optional two-integer tuple that represents the origin of the text on the splash screen image. The
            origin of the text is its lower left corner. A unit in the respective coordinate system is a pixel of the
            image, its origin lies in the top left corner of the image. This parameter also acts like a switch for
            the text feature. If omitted, no text will be displayed on the splash screen. This text will be used to
            show textual progress in onefile mode.
        :type text_pos: Tuple[int, int]
        :keyword text_size:
            The desired size of the font. If the size argument is a positive number, it is interpreted as a size in
            points. If size is a negative number, its absolute value is interpreted as a size in pixels. Default: ``12``
        :type text_size: int
        :keyword text_font:
            An optional name of a font for the text. This font must be installed on the user system, otherwise the
            system default font is used. If this parameter is omitted, the default font is also used.
        :keyword text_color:
            An optional color for the text. HTML color codes (``'#40e0d0'``) and color names (``'turquoise'``) are
            supported. Default: ``'black'``
            (Windows: the color ``'magenta'`` / ``'#ff00ff'`` is used to indicate transparency, and should not be used)
        :type text_color: str
        :keyword text_default:
            The default text which will be displayed before the extraction starts. Default: ``"Initializing"``
        :type text_default: str
        :keyword full_tk:
            By default Splash bundles only the necessary files for the splash screen (some tk components). This
            options enables adding full tk and making it a requirement, meaning all tk files will be unpacked before
            the splash screen can be started. This is useful during development of the splash screen script.
            Default: ``False``
        :type full_tk: bool
        :keyword minify_script:
            The splash screen is created by executing an Tcl/Tk script. This option enables minimizing the script,
            meaning removing all non essential parts from the script. Default: ``True``
        :keyword name:
            An optional alternative filename for the .res file. If not specified, a name is generated.
        :type name: str
        :keyword script_name:
            An optional alternative filename for the Tcl script, that will be generated. If not specified, a name is
            generated.
        :type script_name: str
        :keyword max_img_size:
            Maximum size of the splash screen image as a tuple. If the supplied image exceeds this limit, it will be
            resized to fit the maximum width (to keep the original aspect ratio). This option can be disabled by
            setting it to None. Default: ``(760, 480)``
        :type max_img_size: Tuple[int, int]
        :keyword always_on_top:
            Force the splashscreen to be always on top of other windows. If disabled, other windows (e.g., from other
            applications) can cover the splash screen by user bringing them to front. This might be useful for
            frozen applications with long startup times. Default: ``True``
        :type always_on_top: bool
        """
        from PyInstaller.config import CONF
        from PyInstaller.utils.hooks.tcl_tk import tcltk_info

        Target.__init__(self)

        # Splash screen is not supported on macOS. It operates in a secondary thread and macOS disallows UI operations
        # in any thread other than main.
        if is_darwin:
            raise SystemExit("ERROR: Splash screen is not supported on macOS.")

        # Ensure tkinter (and thus Tcl/Tk) is available.
        if not tcltk_info.available:
            raise SystemExit(
                "ERROR: Your platform does not support the splash screen feature, since tkinter is not installed. "
                "Please install tkinter and try again."
            )

        # Check if the Tcl/Tk version is supported.
        logger.info("Verifying Tcl/Tk compatibility with splash screen requirements")
        self._check_tcl_tk_compatibility(tcltk_info)

        # Make image path relative to .spec file
        if not os.path.isabs(image_file):
            image_file = os.path.join(CONF['specpath'], image_file)
        image_file = os.path.normpath(image_file)
        if not os.path.exists(image_file):
            raise ValueError("Image file '%s' not found" % image_file)

        # Copy all arguments
        self.image_file = image_file
        self.full_tk = kwargs.get("full_tk", False)
        self.name = kwargs.get("name", None)
        self.script_name = kwargs.get("script_name", None)
        self.minify_script = kwargs.get("minify_script", True)
        self.max_img_size = kwargs.get("max_img_size", (760, 480))

        # text options
        self.text_pos = kwargs.get("text_pos", None)
        self.text_size = kwargs.get("text_size", 12)
        self.text_font = kwargs.get("text_font", "TkDefaultFont")
        self.text_color = kwargs.get("text_color", "black")
        self.text_default = kwargs.get("text_default", "Initializing")

        # always-on-top behavior
        self.always_on_top = kwargs.get("always_on_top", True)

        # Save the generated file separately so that it is not necessary to generate the data again and again
        root = os.path.splitext(self.tocfilename)[0]
        if self.name is None:
            self.name = root + '.res'
        if self.script_name is None:
            self.script_name = root + '_script.tcl'

        # Internal variables
        # Store path to _tkinter extension module, so that guts check can detect if the path changed for some reason.
        self._tkinter_file = tcltk_info.tkinter_extension_file

        # Calculated / analysed values
        self.uses_tkinter = self._uses_tkinter(self._tkinter_file, binaries)
        logger.debug("Program uses tkinter: %r", self.uses_tkinter)
        self.script = self.generate_script()
        self.tcl_lib = tcltk_info.tcl_shared_library  # full path to shared library
        self.tk_lib = tcltk_info.tk_shared_library

        assert self.tcl_lib is not None
        assert self.tk_lib is not None

        logger.debug("Using Tcl shared library: %r", self.tcl_lib)
        logger.debug("Using Tk shared library: %r", self.tk_lib)

        self.splash_requirements = set([
            # NOTE: the implicit assumption here is that Tcl and Tk shared library are collected into top-level
            # application directory, which, at tme moment, is true in practically all cases.
            os.path.basename(self.tcl_lib),
            os.path.basename(self.tk_lib),
            # The list of requirements below is based on the current implementation of splash screen script. If you want
            # to extend the splash screen functionality and run into Tcl/Tk errors, chances are that additional Tk
            # components need to be added here.
            #
            # NOTE: these paths use the *destination* layout for Tcl/Tk scripts, which uses unversioned tcl and tk
            # directories (see `PyInstaller.utils.hooks.tcl_tk.collect_tcl_tk_files`).
            os.path.join(tcltk_info.TCL_ROOTNAME, "init.tcl"),
            # Core Tk
            os.path.join(tcltk_info.TK_ROOTNAME, "license.terms"),
            os.path.join(tcltk_info.TK_ROOTNAME, "text.tcl"),
            os.path.join(tcltk_info.TK_ROOTNAME, "tk.tcl"),
            # Used for customizable font
            os.path.join(tcltk_info.TK_ROOTNAME, "ttk", "ttk.tcl"),
            os.path.join(tcltk_info.TK_ROOTNAME, "ttk", "fonts.tcl"),
            os.path.join(tcltk_info.TK_ROOTNAME, "ttk", "cursors.tcl"),
            os.path.join(tcltk_info.TK_ROOTNAME, "ttk", "utils.tcl"),
        ])

        if tcltk_info.tk_version >= (9, 0):
            self.splash_requirements.update([
                os.path.join(tcltk_info.TK_ROOTNAME, "scaling.tcl"),
                os.path.join(tcltk_info.TK_ROOTNAME, "tclIndex"),  # required for auto-load of scaling.tcl
            ])

        logger.info("Collect Tcl/Tk data files for the splash screen")
        tcltk_tree = tcltk_info.data_files  # 3-element tuple TOC
        if self.full_tk:
            # The user wants a full copy of Tk, so make all Tk files a requirement.
            self.splash_requirements.update(entry[0] for entry in tcltk_tree)

        # Scan for binary dependencies of the Tcl/Tk shared libraries, and add them to `binaries` TOC list (which
        # should really be called `dependencies` as it is not limited to binaries. But it is too late now, and
        # existing spec files depend on this naming). We specify these binary dependencies (which include the
        # Tcl and Tk shared libraries themselves) even if the user's program uses tkinter and they would be collected
        # anyway; let the collection mechanism deal with potential duplicates.
        tcltk_libs = [(os.path.basename(src_name), src_name, 'BINARY') for src_name in (self.tcl_lib, self.tk_lib)]
        self.binaries = bindepend.binary_dependency_analysis(tcltk_libs)

        # Put all shared library dependencies in `splash_requirements`, so they are made available in onefile mode.
        self.splash_requirements.update(entry[0] for entry in self.binaries)

        # If the user's program does not use tkinter, add resources from Tcl/Tk tree to the dependencies list.
        # Do so only for the resources that are part of splash requirements.
        if not self.uses_tkinter:
            self.binaries.extend(entry for entry in tcltk_tree if entry[0] in self.splash_requirements)

        # Check if all requirements were found.
        collected_files = set(entry[0] for entry in (binaries + datas + self.binaries))

        def _filter_requirement(filename):
            if filename not in collected_files:
                # Item is not bundled, so warn the user about it. This actually may happen on some tkinter installations
                # that are missing the license.terms file - as this file has no effect on operation of splash screen,
                # suppress the warning for it.
                if os.path.basename(filename) == 'license.terms':
                    return False

                logger.warning(
                    "The local Tcl/Tk installation is missing the file %s. The behavior of the splash screen is "
                    "therefore undefined and may be unsupported.", filename
                )
                return False
            return True

        # Remove all files which were not found.
        self.splash_requirements = set(filter(_filter_requirement, self.splash_requirements))

        logger.debug("Splash Requirements: %s", self.splash_requirements)

        # On AIX, the Tcl and Tk shared libraries might in fact be ar archives with shared object inside it, and need to
        # be `dlopen`'ed with full name (for example, `libtcl.a(libtcl.so.8.6)` and `libtk.a(libtk.so.8.6)`. So if the
        # library's suffix is .a, adjust the name accordingly, assuming fixed format for the shared object name.
        # Adjust the names at the end of this method, because preceding steps use `self.tcl_lib` and `self.tk_lib` for
        # filesystem-based operations and need the original filenames.
        if is_aix:
            _, ext = os.path.splitext(self.tcl_lib)
            if ext == '.a':
                tcl_major, tcl_minor = tcltk_info.tcl_version
                self.tcl_lib += f"(libtcl.so.{tcl_major}.{tcl_minor})"
            _, ext = os.path.splitext(self.tk_lib)
            if ext == '.a':
                tk_major, tk_minor = tcltk_info.tk_version
                self.tk_lib += f"(libtk.so.{tk_major}.{tk_minor})"

        self.__postinit__()

    _GUTS = (
        # input parameters
        ('image_file', _check_guts_eq),
        ('name', _check_guts_eq),
        ('script_name', _check_guts_eq),
        ('text_pos', _check_guts_eq),
        ('text_size', _check_guts_eq),
        ('text_font', _check_guts_eq),
        ('text_color', _check_guts_eq),
        ('text_default', _check_guts_eq),
        ('always_on_top', _check_guts_eq),
        ('full_tk', _check_guts_eq),
        ('minify_script', _check_guts_eq),
        ('max_img_size', _check_guts_eq),
        # calculated/analysed values
        ('uses_tkinter', _check_guts_eq),
        ('script', _check_guts_eq),
        ('tcl_lib', _check_guts_eq),
        ('tk_lib', _check_guts_eq),
        ('splash_requirements', _check_guts_eq),
        ('binaries', _check_guts_toc),
        # internal value
        # Check if the tkinter installation changed. This is theoretically possible if someone uses two different python
        # installations of the same version.
        ('_tkinter_file', _check_guts_eq),
    )

    def _check_guts(self, data, last_build):
        if Target._check_guts(self, data, last_build):
            return True

        # Check if the image has been modified.
        if misc.mtime(self.image_file) > last_build:
            logger.info("Building %s because file %s changed", self.tocbasename, self.image_file)
            return True

        return False

    def assemble(self):
        logger.info("Building Splash %s", self.name)

        # Check if PIL/pillow is available.
        try:
            from PIL import Image as PILImage
        except ImportError:
            PILImage = None

        # Required to pass tcltk_info.TK_ROOTNAME to SplashWriter
        from PyInstaller.utils.hooks.tcl_tk import tcltk_info

        # Function to resize a given image to fit into the area defined by max_img_size.
        def _resize_image(_image, _orig_size):
            if PILImage:
                _w, _h = _orig_size
                _ratio_w = self.max_img_size[0] / _w
                if _ratio_w < 1:
                    # Image width exceeds limit
                    _h = int(_h * _ratio_w)
                    _w = self.max_img_size[0]

                _ratio_h = self.max_img_size[1] / _h
                if _ratio_h < 1:
                    # Image height exceeds limit
                    _w = int(_w * _ratio_h)
                    _h = self.max_img_size[1]

                # If a file is given it will be open
                if isinstance(_image, PILImage.Image):
                    _img = _image
                else:
                    _img = PILImage.open(_image)
                _img_resized = _img.resize((_w, _h))

                # Save image into a stream
                _image_stream = io.BytesIO()
                _img_resized.save(_image_stream, format='PNG')
                _img.close()
                _img_resized.close()
                _image_data = _image_stream.getvalue()
                logger.info("Resized image %s from dimensions %r to (%d, %d)", self.image_file, _orig_size, _w, _h)
                return _image_data
            else:
                raise ValueError(
                    "The splash image dimensions (w: %d, h: %d) exceed max_img_size (w: %d, h:%d), but the image "
                    "cannot be resized due to missing PIL.Image! Either install the Pillow package, adjust the "
                    "max_img_size, or use an image of compatible dimensions." %
                    (_orig_size[0], _orig_size[1], self.max_img_size[0], self.max_img_size[1])
                )

        # Open image file
        image_file = open(self.image_file, 'rb')

        # Check header of the file to identify it
        if image_file.read(8) == b'\x89PNG\r\n\x1a\n':
            # self.image_file is a PNG file
            image_file.seek(16)
            img_size = (struct.unpack("!I", image_file.read(4))[0], struct.unpack("!I", image_file.read(4))[0])

            if img_size > self.max_img_size:
                # The image exceeds the maximum image size, so resize it
                image = _resize_image(self.image_file, img_size)
            else:
                image = os.path.abspath(self.image_file)
        elif PILImage:
            # Pillow is installed, meaning the image can be converted automatically
            img = PILImage.open(self.image_file, mode='r')

            if img.size > self.max_img_size:
                image = _resize_image(img, img.size)
            else:
                image_data = io.BytesIO()
                img.save(image_data, format='PNG')
                img.close()
                image = image_data.getvalue()
            logger.info("Converted image %s to PNG format", self.image_file)
        else:
            raise ValueError(
                "The image %s needs to be converted to a PNG file, but PIL.Image is not available! Either install the "
                "Pillow package, or use a PNG image for you splash screen." % (self.image_file,)
            )

        image_file.close()

        SplashWriter(
            self.name,
            self.splash_requirements,
            os.path.basename(self.tcl_lib),  # tcl86t.dll
            os.path.basename(self.tk_lib),  # tk86t.dll
            tcltk_info.TCL_ROOTNAME,
            tcltk_info.TK_ROOTNAME,
            image,
            self.script
        )

    @staticmethod
    def _check_tcl_tk_compatibility(tcltk_info):
        tcl_version = tcltk_info.tcl_version  # (major, minor) tuple
        tk_version = tcltk_info.tk_version

        if is_darwin and tcltk_info.is_macos_system_framework:
            # Outdated Tcl/Tk 8.5 system framework is not supported.
            raise SystemExit(
                "ERROR: The splash screen feature does not support macOS system framework version of Tcl/Tk."
            )

        # Test if tcl/tk version is supported
        if tcl_version < (8, 6) or tk_version < (8, 6):
            logger.warning(
                "The installed Tcl/Tk (%d.%d / %d.%d) version might not work with the splash screen feature of the "
                "bootloader, which was tested against Tcl/Tk 8.6", *tcl_version, *tk_version
            )

        # This should be impossible, since tcl/tk is released together with the same version number, but just in case
        if tcl_version != tk_version:
            logger.warning(
                "The installed version of Tcl (%d.%d) and Tk (%d.%d) do not match. PyInstaller is tested against "
                "matching versions", *tcl_version, *tk_version
            )

        # Ensure that Tcl is built with multi-threading support.
        if not tcltk_info.tcl_threaded:
            # This is a feature breaking problem, so exit.
            raise SystemExit(
                "ERROR: The installed Tcl version is not threaded. PyInstaller only supports the splash screen "
                "using threaded Tcl."
            )

        # Ensure that Tcl and Tk shared libraries are available
        if tcltk_info.tcl_shared_library is None or tcltk_info.tk_shared_library is None:
            message = \
                "ERROR: Could not determine the path to Tcl and/or Tk shared library, " \
                "which are required for splash screen."
            if not tcltk_info.tkinter_extension_file:
                message += (
                    " The _tkinter module appears to be a built-in, which likely means that python was built with "
                    "statically-linked Tcl/Tk libraries and is incompatible with splash screen."
                )
            raise SystemExit(message)

    def generate_script(self):
        """
        Generate the script for the splash screen.

        If minify_script is True, all unnecessary parts will be removed.
        """
        d = {}
        if self.text_pos is not None:
            logger.debug("Add text support to splash screen")
            d.update({
                'pad_x': self.text_pos[0],
                'pad_y': self.text_pos[1],
                'color': self.text_color,
                'font': self.text_font,
                'font_size': self.text_size,
                'default_text': self.text_default,
            })
        script = splash_templates.build_script(text_options=d, always_on_top=self.always_on_top)

        if self.minify_script:
            # Remove any documentation, empty lines and unnecessary spaces
            script = '\n'.join(
                line for line in map(lambda line: line.strip(), script.splitlines())
                if not line.startswith('#')  # documentation
                and line  # empty lines
            )
            # Remove unnecessary spaces
            script = re.sub(' +', ' ', script)

        # Write script to disk, so that it is transparent to the user what script is executed.
        with open(self.script_name, "w", encoding="utf-8") as script_file:
            script_file.write(script)
        return script

    @staticmethod
    def _uses_tkinter(tkinter_file, binaries):
        # Test for _tkinter extension instead of tkinter module, because user might use a different wrapping library for
        # Tk. Use `pathlib.PurePath` in comparisons to account for case normalization and separator normalization.
        tkinter_file = pathlib.PurePath(tkinter_file)
        for dest_name, src_name, typecode in binaries:
            if pathlib.PurePath(src_name) == tkinter_file:
                return True
        return False



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Functional tests for the Python Imaging Library (PIL).

Note that the original unmaintained PIL has been obsoleted by the PIL-compatible fork Pillow,
which retains the same Python package `PIL`.
"""

import pathlib

import pytest

from PyInstaller.utils.tests import importorskip
from PyInstaller.utils.hooks import can_import_module

# All tests in this file require PIL
pytestmark = importorskip('PIL')


# Functional test that tries to convert a .tiff image to a .png
def test_pil_image_conversion(pyi_builder, tmp_path):
    pyi_builder.test_source(
        """
        import sys
        import os

        import PIL.Image

        if len(sys.argv) != 3:
            print(f"use: {sys.argv[0]} <input-filename> <output-filename>")
            raise SystemExit(1)

        input_file = sys.argv[1]
        output_file = sys.argv[2]

        image = PIL.Image.open(input_file)
        image.save(output_file)
        """,
        app_args=[
            str(pathlib.Path(__file__).parent / 'data' / 'PIL_images' / 'tinysample.tiff'),
            str(tmp_path / 'converted_tinysample.png'),
        ],
    )


@importorskip('PyQt5')
def test_pil_pyqt5(pyi_builder):
    # hook-PIL is excluding PyQt5, but is must still be included since it is imported elsewhere.
    # Also see issue #1584.
    pyi_builder.test_source("""
        import PyQt5
        import PIL
        import PIL.ImageQt
        """)


def test_pil_plugins(pyi_builder):
    pyi_builder.test_source(
        """
        # Verify packaging of PIL.Image.
        from PIL.Image import frombytes
        print(frombytes)

        # PIL import hook should bundle all available PIL plugins. Verify that plugins are collected.
        from PIL import Image
        Image.init()
        MIN_PLUG_COUNT = 7  # Without all plugins the count is usually 6.
        plugins = list(Image.SAVE.keys())
        plugins.sort()
        if len(plugins) < MIN_PLUG_COUNT:
            raise SystemExit('No PIL image plugins were collected!')
        else:
            print('PIL supported image formats: %s' % plugins)
        """
    )


# The tkinter module may be available for import, but not actually importable due to missing shared libraries.
# Therefore, we need to use `can_import_module`-based skip decorator instead of `@importorskip`.
@pytest.mark.skipif(not can_import_module("tkinter"), reason="tkinter cannot be imported.")
def test_pil_no_tkinter(pyi_builder):
    """
    Ensure that the Tkinter package excluded by `PIL` package hooks is unimportable by frozen applications explicitly
    importing only the latter.
    """

    pyi_builder.test_source(
        """
        import PIL.Image

        # Dynamically importing the Tkinter package should fail with an "ImportError", implying "PIL" package hooks
        # successfully excluded Tkinter. To prevent PyInstaller from parsing this import and thus freezing this
        # extension with this test, this import is dynamic.
        try:
            __import__('tkinter')
            raise SystemExit('ERROR: Module tkinter is bundled.')
        except ImportError:
            pass

        # Dynamically importing the "_tkinter" shared library should also fail.
        try:
            __import__('_tkinter')
            raise SystemExit('ERROR: Module _tkinter is bundled.')
        except ImportError:
            pass
        """
    )


# The tkinter module may be available for import, but not actually importable due to missing shared libraries.
# Therefore, we need to use `can_import_module`-based skip decorator instead of `@importorskip`.
@pytest.mark.skipif(not can_import_module("tkinter"), reason="tkinter cannot be imported.")
def test_pil_tkinter(pyi_builder):
    """
    Ensure that the Tkinter package excluded by `PIL` package hooks is importable by frozen applications explicitly
    importing both.

    == See Also ==

    * PyInstaller [issue #1584](https://github.com/pyinstaller/pyinstaller/issues/1584).
    """

    pyi_builder.test_source(
        """
        import PIL.Image

        # Statically importing the Tkinter package should succeed, implying this importation successfully overrode
        # the exclusion of this package requested by "PIL" package hooks. To ensure PyInstaller parses this import
        # and freezes this package with this test, this import is static.
        try:
            import tkinter
        except ImportError:
            raise SystemExit('ERROR: Module tkinter is NOT bundled.')
        """
    )


@importorskip('matplotlib')
def test_pil_no_matplotlib(pyi_builder):
    """
    Ensure that using PIL.Image does not pull in `matplotlib` when the latter is not explicitly imported by the program.
    The import chain in question,
    PIL.Image -> PIL -> PIL.ImageShow -> IPython -> matplotlib_inline.backend_inline -> matplotlib,
    should be broken by the PIL hook excluding IPython.
    """

    pyi_builder.test_source(
        """
        import PIL.Image

        # Use dynamic import of matplotlib to prevent PyInstaller from picking up the import.
        try:
            __import__('matplotlib')
            raise SystemExit('ERROR: matplotlib is bundled.')
        except ImportError:
            pass
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pytest

from PyInstaller.compat import is_nogil

# All tests in this file require freethreading-enabled python build (i.e., built with Py_GIL_DISABLED).
pytestmark = pytest.mark.skipif(not is_nogil, reason="Requires freethreading-enabled python build.")


# Test that GIL can be explicitly disabled or re-enabled via `X gil=0` and `X gil=1` python option.
@pytest.mark.parametrize('gil_value', [0, 1], ids=['enabled', 'disabled'])
def test_freethreading_explicit_gil_setting(pyi_builder, gil_value):
    pyi_builder.test_source(
        """
        import sys

        # Expected status is passed via first argument, as 0/1 integer.
        expected_status = bool(int(sys.argv[1]))

        status = sys._is_gil_enabled()
        print("GIL status: {status}")
        if status != expected_status:
            raise ValueError(f"Unexpected GIL status: {status}; expected {expected_status}")
        """,
        pyi_args=['--python-option', f'X gil={gil_value}'],
        app_args=[str(gil_value)],
    )



import inspect


def test_call_chain():
    modules = []
    for idx, frame_info in enumerate(inspect.stack()):
        module = inspect.getmodule(frame_info[0])
        assert module, f"Failed to get module for frame #{idx}: {frame_info}"
        modules.append(module)
    return modules



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, stat
from waflib import Utils, Build, Node

STRONGEST = True
Build.SAVED_ATTRS.append('hashes_md5_tstamp')


def h_file(self):
    filename = self.abspath()
    st = os.stat(filename)
    cache = self.ctx.hashes_md5_tstamp
    if filename in cache and cache[filename][0] == st.st_mtime:
        return cache[filename][1]
    if STRONGEST:
        ret = Utils.h_file(filename)
    else:
        if stat.S_ISDIR(st[stat.ST_MODE]):
            raise IOError('Not a file')
        ret = Utils.md5(str((st.st_mtime, st.st_size)).encode()).digest()
    cache[filename] = (st.st_mtime, ret)
    return ret


h_file.__doc__ = Node.Node.h_file.__doc__
Node.Node.h_file = h_file



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This script is used by multiple tests.
# It checks that various paths set by the bootloader are usable filenames.

import sys
import os

import win32api


def check_shortpathname(fn):
    lfn = win32api.GetLongPathNameW(fn)
    fn = os.path.normcase(fn)
    lfn = os.path.normcase(lfn)
    if lfn != fn:
        print("ShortPathName: Expected %s, got %s" % (fn, lfn))
        raise SystemExit(-1)


print("sys.executable:", ascii(sys.executable))

if not os.path.exists(sys.executable):
    raise SystemExit("sys.executable does not exist.")
check_shortpathname(sys.executable)

print("sys.argv[0]:", ascii(sys.argv[0]))

if not os.path.exists(sys.argv[0]):
    raise SystemExit("sys.argv[0] does not exist.")
check_shortpathname(sys.argv[0])

print("sys._MEIPASS:", ascii(sys._MEIPASS))

if not os.path.exists(sys._MEIPASS):
    raise SystemExit("sys._MEIPASS does not exist.")
tmp = os.path.normcase(win32api.GetTempPath())
if os.path.normcase(win32api.GetLongPathNameW(tmp)) == tmp:
    # Test only if TempPath is not a short path. This might for example happen if
    # TMP=c:\users\runner~1\appdata\local\temp, due to  username being too long.
    check_shortpathname(sys._MEIPASS)



# -----------------------------------------------------------------------------
# Copyright (c) 2014-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------
import socket

try:
    import BaseHTTPServer
    import SimpleHTTPServer
except ImportError:
    import http.server as BaseHTTPServer
    import http.server as SimpleHTTPServer

import os
import ssl
import sys
import threading
import time

import requests
"""
Note: to re-create the server.pem file use the following commands:

cd /path/to/pyinstaller.git/tests/functional
openssl req -new -x509 -keyout data/requests/server.pem \
    -text -out data/requests/server.pem -days 36500 \
    -nodes -config data/requests/openssl.conf
"""

if getattr(sys, 'frozen', False):
    # We are running in a |PyInstaller| bundle.
    basedir = sys._MEIPASS
else:
    # We are running in a normal Python environment.
    basedir = os.path.dirname(__file__)

SERVER_CERT = os.path.join(basedir, "server.pem")

if not os.path.exists(SERVER_CERT):
    raise SystemExit('Certificate-File %s is missing' % SERVER_CERT)


def main():

    SERVER_PORT = 8443
    httpd = None

    # Since unit tests run in parallel, the port may be in use, so retry creating the server while incrementing
    # the port number.
    while SERVER_PORT < 8493:  # Max 50 retries
        try:
            # SSL server copied from here: http://www.piware.de/2011/01/creating-an-https-server-in-python/
            httpd = BaseHTTPServer.HTTPServer(('localhost', SERVER_PORT), SimpleHTTPServer.SimpleHTTPRequestHandler)
        except socket.error as e:
            if e.errno == 98:  # Address in use
                SERVER_PORT += 1
                continue
            else:
                # Some other socket.error
                raise
        else:
            # Success
            break
    else:
        # Did not break from loop, so we ran out of retries
        assert False, "Could not bind server port: all ports in use."

    # Wrap the server socket using SSL context. The old `ssl.wrap_socket` was deprecated in python 3.7 and removed
    # in python 3.12.
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    ssl_context.load_cert_chain(certfile=SERVER_CERT, keyfile=None)
    httpd.socket = ssl_context.wrap_socket(
        httpd.socket,
        server_side=True,
    )

    def ssl_server():
        httpd.serve_forever()

    # Start the SSL server
    thread = threading.Thread(target=ssl_server)
    thread.daemon = True
    thread.start()

    # Wait a bit for the server to start
    time.sleep(1)

    # Use requests to get a page from the server
    requests.get("https://localhost:{}".format(SERVER_PORT), verify=SERVER_CERT)
    # requests.get("https://github.com")


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import heapq, traceback
try:
    from queue import Queue, PriorityQueue
except ImportError:
    from Queue import Queue
    try:
        from Queue import PriorityQueue
    except ImportError:

        class PriorityQueue(Queue):
            def _init(self, maxsize):
                self.maxsize = maxsize
                self.queue = []

            def _put(self, item):
                heapq.heappush(self.queue, item)

            def _get(self):
                return heapq.heappop(self.queue)


from waflib import Utils, Task, Errors, Logs

GAP = 5


class PriorityTasks(object):
    def __init__(self):
        self.lst = []

    def __len__(self):
        return len(self.lst)

    def __iter__(self):
        return iter(self.lst)

    def __str__(self):
        return 'PriorityTasks: [%s]' % '\n  '.join(str(x) for x in self.lst)

    def clear(self):
        self.lst = []

    def append(self, task):
        heapq.heappush(self.lst, task)

    def appendleft(self, task):
        heapq.heappush(self.lst, task)

    def pop(self):
        return heapq.heappop(self.lst)

    def extend(self, lst):
        if self.lst:
            for x in lst:
                self.append(x)
        else:
            if isinstance(lst, list):
                self.lst = lst
                heapq.heapify(lst)
            else:
                self.lst = lst.lst


class Consumer(Utils.threading.Thread):
    def __init__(self, spawner, task):
        Utils.threading.Thread.__init__(self)
        self.task = task
        self.spawner = spawner
        self.daemon = True
        self.start()

    def run(self):
        try:
            if not self.spawner.master.stop:
                self.spawner.master.process_task(self.task)
        finally:
            self.spawner.sem.release()
            self.spawner.master.out.put(self.task)
            self.task = None
            self.spawner = None


class Spawner(Utils.threading.Thread):
    def __init__(self, master):
        Utils.threading.Thread.__init__(self)
        self.master = master
        self.sem = Utils.threading.Semaphore(master.numjobs)
        self.daemon = True
        self.start()

    def run(self):
        try:
            self.loop()
        except Exception:
            pass

    def loop(self):
        master = self.master
        while 1:
            task = master.ready.get()
            self.sem.acquire()
            if not master.stop:
                task.log_display(task.generator.bld)
            Consumer(self, task)


class Parallel(object):
    def __init__(self, bld, j=2):
        self.numjobs = j
        self.bld = bld
        self.outstanding = PriorityTasks()
        self.postponed = PriorityTasks()
        self.incomplete = set()
        self.ready = PriorityQueue(0)
        self.out = Queue(0)
        self.count = 0
        self.processed = 0
        self.stop = False
        self.error = []
        self.biter = None
        self.dirty = False
        self.revdeps = Utils.defaultdict(set)
        self.spawner = None
        if self.numjobs > 1:
            self.spawner = Spawner(self)

    def get_next_task(self):
        if not self.outstanding:
            return None
        return self.outstanding.pop()

    def postpone(self, tsk):
        self.postponed.append(tsk)

    def refill_task_list(self):
        while self.count > self.numjobs * GAP:
            self.get_out()
        while not self.outstanding:
            if self.count:
                self.get_out()
                if self.outstanding:
                    break
            elif self.postponed:
                try:
                    cond = self.deadlock == self.processed
                except AttributeError:
                    pass
                else:
                    if cond:
                        lst = []
                        for tsk in self.postponed:
                            deps = [id(x) for x in tsk.run_after if not x.hasrun]
                            lst.append('%s\t-> %r' % (repr(tsk), deps))
                            if not deps:
                                lst.append('\n  task %r dependencies are done, check its *runnable_status*?' % id(tsk))
                        raise Errors.WafError('Deadlock detected: check the task build order%s' % ''.join(lst))
                self.deadlock = self.processed
            if self.postponed:
                self.outstanding.extend(self.postponed)
                self.postponed.clear()
            elif not self.count:
                if self.incomplete:
                    for x in self.incomplete:
                        for k in x.run_after:
                            if not k.hasrun:
                                break
                        else:
                            self.incomplete.remove(x)
                            self.outstanding.append(x)
                            break
                    else:
                        if self.stop or self.error:
                            break
                        raise Errors.WafError('Broken revdeps detected on %r' % self.incomplete)
                else:
                    tasks = next(self.biter)
                    ready, waiting = self.prio_and_split(tasks)
                    self.outstanding.extend(ready)
                    self.incomplete.update(waiting)
                    self.total = self.bld.total()
                    break

    def add_more_tasks(self, tsk):
        if getattr(tsk, 'more_tasks', None):
            more = set(tsk.more_tasks)
            groups_done = set()

            def iteri(a, b):
                for x in a:
                    yield x
                for x in b:
                    yield x

            for x in iteri(self.outstanding, self.incomplete):
                for k in x.run_after:
                    if isinstance(k, Task.TaskGroup):
                        if k not in groups_done:
                            groups_done.add(k)
                            for j in k.prev & more:
                                self.revdeps[j].add(k)
                    elif k in more:
                        self.revdeps[k].add(x)
            ready, waiting = self.prio_and_split(tsk.more_tasks)
            self.outstanding.extend(ready)
            self.incomplete.update(waiting)
            self.total += len(tsk.more_tasks)

    def mark_finished(self, tsk):
        def try_unfreeze(x):
            if x in self.incomplete:
                for k in x.run_after:
                    if not k.hasrun:
                        break
                else:
                    self.incomplete.remove(x)
                    self.outstanding.append(x)

        if tsk in self.revdeps:
            for x in self.revdeps[tsk]:
                if isinstance(x, Task.TaskGroup):
                    x.prev.remove(tsk)
                    if not x.prev:
                        for k in x.next:
                            k.run_after.remove(x)
                            try_unfreeze(k)
                        x.next = []
                else:
                    try_unfreeze(x)
            del self.revdeps[tsk]
        if hasattr(tsk, 'semaphore'):
            sem = tsk.semaphore
            try:
                sem.release(tsk)
            except KeyError:
                pass
            else:
                while sem.waiting and not sem.is_locked():
                    x = sem.waiting.pop()
                    self._add_task(x)

    def get_out(self):
        tsk = self.out.get()
        if not self.stop:
            self.add_more_tasks(tsk)
        self.mark_finished(tsk)
        self.count -= 1
        self.dirty = True
        return tsk

    def add_task(self, tsk):
        self.ready.put(tsk)

    def _add_task(self, tsk):
        if hasattr(tsk, 'semaphore'):
            sem = tsk.semaphore
            try:
                sem.acquire(tsk)
            except IndexError:
                sem.waiting.add(tsk)
                return
        self.count += 1
        self.processed += 1
        if self.numjobs == 1:
            tsk.log_display(tsk.generator.bld)
            try:
                self.process_task(tsk)
            finally:
                self.out.put(tsk)
        else:
            self.add_task(tsk)

    def process_task(self, tsk):
        tsk.process()
        if tsk.hasrun != Task.SUCCESS:
            self.error_handler(tsk)

    def skip(self, tsk):
        tsk.hasrun = Task.SKIPPED
        self.mark_finished(tsk)

    def cancel(self, tsk):
        tsk.hasrun = Task.CANCELED
        self.mark_finished(tsk)

    def error_handler(self, tsk):
        if not self.bld.keep:
            self.stop = True
        self.error.append(tsk)

    def task_status(self, tsk):
        try:
            return tsk.runnable_status()
        except Exception:
            self.processed += 1
            tsk.err_msg = traceback.format_exc()
            if not self.stop and self.bld.keep:
                self.skip(tsk)
                if self.bld.keep == 1:
                    if Logs.verbose > 1 or not self.error:
                        self.error.append(tsk)
                    self.stop = True
                else:
                    if Logs.verbose > 1:
                        self.error.append(tsk)
                return Task.EXCEPTION
            tsk.hasrun = Task.EXCEPTION
            self.error_handler(tsk)
            return Task.EXCEPTION

    def start(self):
        self.total = self.bld.total()
        while not self.stop:
            self.refill_task_list()
            tsk = self.get_next_task()
            if not tsk:
                if self.count:
                    continue
                else:
                    break
            if tsk.hasrun:
                self.processed += 1
                continue
            if self.stop:
                break
            st = self.task_status(tsk)
            if st == Task.RUN_ME:
                self._add_task(tsk)
            elif st == Task.ASK_LATER:
                self.postpone(tsk)
            elif st == Task.SKIP_ME:
                self.processed += 1
                self.skip(tsk)
                self.add_more_tasks(tsk)
            elif st == Task.CANCEL_ME:
                if Logs.verbose > 1:
                    self.error.append(tsk)
                self.processed += 1
                self.cancel(tsk)
        while self.error and self.count:
            self.get_out()
        self.ready.put(None)
        if not self.stop:
            assert not self.count
            assert not self.postponed
            assert not self.incomplete

    def prio_and_split(self, tasks):
        for x in tasks:
            x.visited = 0
        reverse = self.revdeps
        groups_done = set()
        for x in tasks:
            for k in x.run_after:
                if isinstance(k, Task.TaskGroup):
                    if k not in groups_done:
                        groups_done.add(k)
                        for j in k.prev:
                            reverse[j].add(k)
                else:
                    reverse[k].add(x)

        def visit(n):
            if isinstance(n, Task.TaskGroup):
                return sum(visit(k) for k in n.next)
            if n.visited == 0:
                n.visited = 1
                if n in reverse:
                    rev = reverse[n]
                    n.prio_order = n.tree_weight + len(rev) + sum(visit(k) for k in rev)
                else:
                    n.prio_order = n.tree_weight
                n.visited = 2
            elif n.visited == 1:
                raise Errors.WafError('Dependency cycle found!')
            return n.prio_order

        for x in tasks:
            if x.visited != 0:
                continue
            try:
                visit(x)
            except Errors.WafError:
                self.debug_cycles(tasks, reverse)
        ready = []
        waiting = []
        for x in tasks:
            for k in x.run_after:
                if not k.hasrun:
                    waiting.append(x)
                    break
            else:
                ready.append(x)
        return (ready, waiting)

    def debug_cycles(self, tasks, reverse):
        tmp = {}
        for x in tasks:
            tmp[x] = 0

        def visit(n, acc):
            if isinstance(n, Task.TaskGroup):
                for k in n.next:
                    visit(k, acc)
                return
            if tmp[n] == 0:
                tmp[n] = 1
                for k in reverse.get(n, []):
                    visit(k, [n] + acc)
                tmp[n] = 2
            elif tmp[n] == 1:
                lst = []
                for tsk in acc:
                    lst.append(repr(tsk))
                    if tsk is n:
                        break
                raise Errors.WafError('Task dependency cycle in "run_after" constraints: %s' % ''.join(lst))

        for x in tasks:
            visit(x, [])



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: Apache-2.0
#-----------------------------------------------------------------------------


def _pyi_rthook():
    import inspect
    import os
    import sys
    import zipfile

    # Use sys._MEIPASS with normalized path component separator. This is necessary on some platforms (i.e., msys2/mingw
    # python on Windows), because we use string comparisons on the paths.
    SYS_PREFIX = os.path.normpath(sys._MEIPASS)
    BASE_LIBRARY = os.path.join(SYS_PREFIX, "base_library.zip")

    # Obtain the list of modules in base_library.zip, so we can use it in our `_pyi_getsourcefile` implementation.
    def _get_base_library_files(filename):
        # base_library.zip might not exit
        if not os.path.isfile(filename):
            return set()

        with zipfile.ZipFile(filename, 'r') as zf:
            namelist = zf.namelist()

        return set(os.path.normpath(entry) for entry in namelist)

    base_library_files = _get_base_library_files(BASE_LIBRARY)

    # Provide custom implementation of inspect.getsourcefile() for frozen applications that properly resolves relative
    # filenames obtained from object (e.g., inspect stack-frames). See #5963.
    #
    # Although we are overriding `inspect.getsourcefile` function, we are NOT trying to resolve source file here!
    # The main purpose of this implementation is to properly resolve relative file names obtained from `co_filename`
    # attribute of code objects (which are, in turn, obtained from in turn are obtained from `frame` and `traceback`
    # objects). PyInstaller strips absolute paths from `co_filename` when collecting modules, as the original absolute
    # paths are not portable/relocatable anyway. The `inspect` module tries to look up the module that corresponds to
    # the code object by comparing modules' `__file__` attribute to the value of `co_filename`. Therefore, our override
    # needs to resolve the relative file names (usually having a .py suffix) into absolute module names (which, in the
    # frozen application, usually have .pyc suffix).
    #
    # The `inspect` module retrieves the actual source code using `linecache.getlines()`. If the passed source filename
    # does not exist, the underlying implementation end up resolving the module, and obtains the source via loader's
    # `get_source` method. So for modules in the PYZ archive, it ends up calling `get_source` implementation on our
    # `PyiFrozenLoader`. For modules in `base_library.zip`, it ends up calling `get_source` on python's own
    # `zipimport.zipimporter`; to properly handle out-of-zip source files, we therefore need to monkey-patch
    # `get_source` with our own override that translates the in-zip .pyc filename into out-of-zip .py file location
    # and loads the source (this override is done in `pyimod02_importers` module).
    #
    # The above-described fallback takes place if the .pyc file does not exist on filesystem - if this ever becomes
    # a problem, we could consider monkey-patching `linecache.updatecache` (and possibly `checkcache`) to translate
    # .pyc paths in `sys._MEIPASS` and `base_library.zip` into .py paths in `sys._MEIPASS` before calling the original
    # implementation.
    _orig_inspect_getsourcefile = inspect.getsourcefile

    def _pyi_getsourcefile(object):
        filename = inspect.getfile(object)
        filename = os.path.normpath(filename)  # Ensure path component separators are normalized.
        if not os.path.isabs(filename):
            # Check if given filename matches the basename of __main__'s __file__.
            main_file = getattr(sys.modules['__main__'], '__file__', None)
            if main_file and filename == os.path.basename(main_file):
                return main_file

            # If the relative filename does not correspond to the frozen entry-point script, convert it to the absolute
            # path in either `sys._MEIPASS/base_library.zip` or `sys._MEIPASS`, whichever applicable.
            #
            # The modules in `sys._MEIPASS/base_library.zip` are handled by python's `zipimport.zipimporter`, and have
            # their __file__ attribute point to the .pyc file in the archive. So we match the behavior, in order to
            # facilitate matching via __file__ attribute and use of loader's `get_source`, as per the earlier comment
            # block.
            #
            # The modules in PYZ archive are handled by our `PyFrozenLoader`, which now sets the module's __file__
            # attribute to point to where .py files would be. Therefore, we can directly merge SYS_PREFIX and filename
            # (and if the source .py file exists, it will be loaded directly from filename, without the intermediate
            # loader look-up).
            pyc_filename = filename + 'c'
            if pyc_filename in base_library_files:
                return os.path.normpath(os.path.join(BASE_LIBRARY, pyc_filename))
            return os.path.normpath(os.path.join(SYS_PREFIX, filename))
        elif filename.startswith(SYS_PREFIX):
            # If filename is already an absolute file path pointing into application's top-level directory, return it
            # as-is and prevent any further processing.
            return filename
        # Use original implementation as a fallback.
        return _orig_inspect_getsourcefile(object)

    inspect.getsourcefile = _pyi_getsourcefile


_pyi_rthook()
del _pyi_rthook



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import fnmatch
import glob
import hashlib
import io
import marshal
import os
import pathlib
import platform
import shutil
import struct
import subprocess
import sys
import types
import zipfile

from PyInstaller import compat
from PyInstaller import log as logging
from PyInstaller.compat import is_aix, is_darwin, is_win, is_linux
from PyInstaller.exceptions import InvalidSrcDestTupleError
from PyInstaller.utils import misc

if is_win:
    from PyInstaller.utils.win32 import versioninfo

if is_darwin:
    import PyInstaller.utils.osx as osxutils

logger = logging.getLogger(__name__)

# -- Helpers for checking guts.
#
# NOTE: by _GUTS it is meant intermediate files and data structures that PyInstaller creates for bundling files and
# creating final executable.


def _check_guts_eq(attr_name, old_value, new_value, last_build):
    """
    Rebuild is required if values differ.
    """
    if old_value != new_value:
        logger.info("Building because %s changed", attr_name)
        return True
    return False


def _check_guts_toc_mtime(attr_name, old_toc, new_toc, last_build):
    """
    Rebuild is required if mtimes of files listed in old TOC are newer than last_build.

    Use this for calculated/analysed values read from cache.
    """
    for dest_name, src_name, typecode in old_toc:
        if misc.mtime(src_name) > last_build:
            logger.info("Building because %s changed", src_name)
            return True
    return False


def _check_guts_toc(attr_name, old_toc, new_toc, last_build):
    """
    Rebuild is required if either TOC content changed or mtimes of files listed in old TOC are newer than last_build.

    Use this for input parameters.
    """
    return _check_guts_eq(attr_name, old_toc, new_toc, last_build) or \
        _check_guts_toc_mtime(attr_name, old_toc, new_toc, last_build)


def destination_name_for_extension(module_name, src_name, typecode):
    """
    Take a TOC entry (dest_name, src_name, typecode) and determine the full destination name for the extension.
    """

    assert typecode == 'EXTENSION'

    # The `module_name` should be the extension's importable module name, such as `psutil._psutil_linux` or
    # `numpy._core._multiarray_umath`. Reconstruct the directory structure from parent package name(s), if any.
    dest_elements = module_name.split('.')

    # We have the base name of the extension file (the last element in the module name), but we do not know the
    # full extension suffix. We can take that from source name; for simplicity, replace the whole base name part.
    src_path = pathlib.Path(src_name)
    dest_elements[-1] = src_path.name

    # Extensions that originate from python's python3.x/lib-dynload directory should be diverted into
    # python3.x/lib-dynload destination directory instead of being collected into top-level application directory.
    # See #5604 for original motivation (using just lib-dynload), and #9204 for extension (using python3.x/lib-dynload).
    if src_path.parent.name == 'lib-dynload':
        python_dir = f'python{sys.version_info.major}.{sys.version_info.minor}'
        if src_path.parent.parent.name == python_dir:
            dest_elements = [python_dir, 'lib-dynload', *dest_elements]

    return os.path.join(*dest_elements)


def process_collected_binary(
    src_name,
    dest_name,
    use_strip=False,
    use_upx=False,
    upx_exclude=None,
    target_arch=None,
    codesign_identity=None,
    entitlements_file=None,
    strict_arch_validation=False
):
    """
    Process the collected binary using strip or UPX (or both), and apply any platform-specific processing. On macOS,
    this rewrites the library paths in the headers, and (re-)signs the binary. On-disk cache is used to avoid processing
    the same binary with same options over and over.

    In addition to given arguments, this function also uses CONF['cachedir'] and CONF['upx_dir'].
    """
    from PyInstaller.config import CONF

    # We need to use cache in the following scenarios:
    #  * extra binary processing due to use of `strip` or `upx`
    #  * building on macOS, where we need to rewrite library paths in binaries' headers and (re-)sign the binaries.
    if not use_strip and not use_upx and not is_darwin:
        return src_name

    # Match against provided UPX exclude patterns.
    upx_exclude = upx_exclude or []
    if use_upx:
        src_path = pathlib.PurePath(src_name)
        for upx_exclude_entry in upx_exclude:
            # pathlib.PurePath.match() matches from right to left, and supports * wildcard, but does not support the
            # "**" syntax for directory recursion. Case sensitivity follows the OS default.
            if src_path.match(upx_exclude_entry):
                logger.info("Disabling UPX for %s due to match in exclude pattern: %s", src_name, upx_exclude_entry)
                use_upx = False
                break

    # Additional automatic disablement rules for UPX and strip.

    # On Windows, avoid using UPX with binaries that have control flow guard (CFG) enabled.
    if use_upx and is_win and versioninfo.pefile_check_control_flow_guard(src_name):
        logger.info('Disabling UPX for %s due to CFG!', src_name)
        use_upx = False

    # Avoid using UPX with Qt plugins, as it strips the data required by the Qt plugin loader.
    if use_upx and misc.is_file_qt_plugin(src_name):
        logger.info('Disabling UPX for %s due to it being a Qt plugin!', src_name)
        use_upx = False

    # On linux, if a binary has an accompanying HMAC or CHK file, avoid modifying it in any way.
    if (use_upx or use_strip) and is_linux:
        src_path = pathlib.Path(src_name)
        hmac_path = src_path.with_name(f".{src_path.name}.hmac")
        chk_path = src_path.with_suffix(".chk")
        if hmac_path.is_file():
            logger.info('Disabling UPX and/or strip for %s due to accompanying .hmac file!', src_name)
            use_upx = use_strip = False
        elif chk_path.is_file():
            logger.info('Disabling UPX and/or strip for %s due to accompanying .chk file!', src_name)
            use_upx = use_strip = False
        del src_path, hmac_path, chk_path

    # Exit early if no processing is required after above rules are applied.
    if not use_strip and not use_upx and not is_darwin:
        return src_name

    # Prepare cache directory path. Cache is tied to python major/minor version, but also to various processing options.
    pyver = f'py{sys.version_info[0]}{sys.version_info[1]}'
    arch = platform.architecture()[0]
    cache_dir = os.path.join(
        CONF['cachedir'],
        f'bincache{use_strip:d}{use_upx:d}{pyver}{arch}',
    )
    if target_arch:
        cache_dir = os.path.join(cache_dir, target_arch)
    if is_darwin:
        # Separate by codesign identity
        if codesign_identity:
            # Compute hex digest of codesign identity string to prevent issues with invalid characters.
            csi_hash = hashlib.sha256(codesign_identity.encode('utf-8'))
            cache_dir = os.path.join(cache_dir, csi_hash.hexdigest())
        else:
            cache_dir = os.path.join(cache_dir, 'adhoc')  # ad-hoc signing
        # Separate by entitlements
        if entitlements_file:
            # Compute hex digest of entitlements file contents
            with open(entitlements_file, 'rb') as fp:
                ef_hash = hashlib.sha256(fp.read())
            cache_dir = os.path.join(cache_dir, ef_hash.hexdigest())
        else:
            cache_dir = os.path.join(cache_dir, 'no-entitlements')
    os.makedirs(cache_dir, exist_ok=True)

    # Load cache index, if available
    cache_index_file = os.path.join(cache_dir, "index.dat")
    try:
        cache_index = misc.load_py_data_struct(cache_index_file)
    except FileNotFoundError:
        cache_index = {}
    except Exception:
        # Tell the user they may want to fix their cache... However, do not delete it for them; if it keeps getting
        # corrupted, we will never find out.
        logger.warning("PyInstaller bincache may be corrupted; use pyinstaller --clean to fix it.")
        raise

    # Look up the file in cache; use case-normalized destination name as identifier.
    cached_id = os.path.normcase(dest_name)
    cached_name = os.path.join(cache_dir, dest_name)
    src_digest = _compute_file_digest(src_name)

    if cached_id in cache_index:
        # If digest matches to the cached digest, return the cached file...
        if src_digest == cache_index[cached_id]:
            return cached_name

        # ... otherwise remove it.
        os.remove(cached_name)

    # Ensure parent path exists
    os.makedirs(os.path.dirname(cached_name), exist_ok=True)

    # Use `shutil.copyfile` to copy the file with default permissions bits, then manually set executable
    # bits. This way, we avoid copying permission bits and metadata from the original file, which might be too
    # restrictive for further processing (read-only permissions, immutable flag on FreeBSD, and so on).
    shutil.copyfile(src_name, cached_name)
    os.chmod(cached_name, 0o755)

    # Apply strip
    if use_strip:
        strip_options = []
        if is_darwin:
            # The default strip behavior breaks some shared libraries under macOS.
            strip_options = ["-S"]  # -S = strip only debug symbols.
        elif is_aix:
            # Set -X32_64 flag to have strip transparently process both 32-bit and 64-bit binaries, without user having
            # to set OBJECT_MODE environment variable prior to the build. Also accommodates potential mixed-case
            # scenario, for example a 32-bit utility program being collected into a 64-bit application bundle.
            strip_options = ["-X32_64"]

        cmd = ["strip", *strip_options, cached_name]
        logger.info("Executing: %s", " ".join(cmd))
        try:
            p = subprocess.run(
                cmd,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
                errors='ignore',
                encoding='utf-8',
            )
            logger.debug("Output from strip command:\n%s", p.stdout)
        except subprocess.CalledProcessError as e:
            show_warning = True

            # On AIX, strip utility raises an error when ran against already-stripped binary. Catch the corresponding
            # message (`0654-419 The specified archive file was already stripped.`) and suppress the warning.
            if is_aix and "0654-419" in e.stdout:
                show_warning = False

            if show_warning:
                logger.warning("Failed to run strip on %r!", cached_name, exc_info=True)
                logger.warning("Output from strip command:\n%s", e.stdout)
        except Exception:
            logger.warning("Failed to run strip on %r!", cached_name, exc_info=True)

    # Apply UPX
    if use_upx:
        upx_exe = 'upx'
        upx_dir = CONF['upx_dir']
        if upx_dir:
            upx_exe = os.path.join(upx_dir, upx_exe)

        upx_options = [
            # Do not compress icons, so that they can still be accessed externally.
            '--compress-icons=0',
            # Use LZMA compression.
            '--lzma',
            # Quiet mode.
            '-q',
        ]
        if is_win:
            # Binaries built with Visual Studio 7.1 require --strip-loadconf or they will not compress.
            upx_options.append('--strip-loadconf')

        cmd = [upx_exe, *upx_options, cached_name]
        logger.info("Executing: %s", " ".join(cmd))
        try:
            p = subprocess.run(
                cmd,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
                errors='ignore',
                encoding='utf-8',
            )
            logger.debug("Output from upx command:\n%s", p.stdout)
        except subprocess.CalledProcessError as e:
            logger.warning("Failed to upx strip on %r!", cached_name, exc_info=True)
            logger.warning("Output from upx command:\n%s", e.stdout)
        except Exception:
            logger.warning("Failed to run upx on %r!", cached_name, exc_info=True)

    # On macOS, we need to modify the given binary's paths to the dependent libraries, in order to ensure they are
    # relocatable and always refer to location within the frozen application. Specifically, we make all dependent
    # library paths relative to @rpath, and set @rpath to point to the top-level application directory, relative to
    # the binary's location (i.e., @loader_path).
    #
    # While modifying the headers invalidates existing signatures, we avoid removing them in order to speed things up
    # (and to avoid potential bugs in the codesign utility, like the one reported on macOS 10.13 in #6167).
    # The forced re-signing at the end should take care of the invalidated signatures.
    if is_darwin:
        try:
            osxutils.binary_to_target_arch(cached_name, target_arch, display_name=src_name)
            #osxutils.remove_signature_from_binary(cached_name)  # Disabled as per comment above.
            target_rpath = str(
                pathlib.PurePath('@loader_path', *['..' for level in pathlib.PurePath(dest_name).parent.parts])
            )
            osxutils.set_dylib_dependency_paths(cached_name, target_rpath)
            osxutils.sign_binary(cached_name, codesign_identity, entitlements_file)
        except osxutils.InvalidBinaryError:
            # Raised by osxutils.binary_to_target_arch when the given file is not a valid macOS binary (for example,
            # a linux .so file; see issue #6327). The error prevents any further processing, so just ignore it.
            pass
        except osxutils.IncompatibleBinaryArchError:
            # Raised by osxutils.binary_to_target_arch when the given file does not contain (all) required arch slices.
            # Depending on the strict validation mode, re-raise or swallow the error.
            #
            # Strict validation should be enabled only for binaries where the architecture *must* match the target one,
            # i.e., the extension modules. Everything else is pretty much a gray area, for example:
            #  * a universal2 extension may have its x86_64 and arm64 slices linked against distinct single-arch/thin
            #    shared libraries
            #  * a collected executable that is launched by python code via a subprocess can be x86_64-only, even though
            #    the actual python code is running on M1 in native arm64 mode.
            if strict_arch_validation:
                raise
            logger.debug("File %s failed optional architecture validation - collecting as-is!", src_name)
        except Exception as e:
            raise SystemError(f"Failed to process binary {cached_name!r}!") from e

    # Update cache index
    cache_index[cached_id] = src_digest
    misc.save_py_data_struct(cache_index_file, cache_index)

    return cached_name


def _compute_file_digest(filename):
    hasher = hashlib.sha1()
    with open(filename, "rb") as fp:
        for chunk in iter(lambda: fp.read(16 * 1024), b""):
            hasher.update(chunk)
    return bytearray(hasher.digest())


def _check_path_overlap(path):
    """
    Check that path does not overlap with WORKPATH or SPECPATH (i.e., WORKPATH and SPECPATH may not start with path,
    which could be caused by a faulty hand-edited specfile).

    Raise SystemExit if there is overlap, return True otherwise
    """
    from PyInstaller.config import CONF
    specerr = 0
    if CONF['workpath'].startswith(path):
        logger.error('Specfile error: The output path "%s" contains WORKPATH (%s)', path, CONF['workpath'])
        specerr += 1
    if CONF['specpath'].startswith(path):
        logger.error('Specfile error: The output path "%s" contains SPECPATH (%s)', path, CONF['specpath'])
        specerr += 1
    if specerr:
        raise SystemExit(
            'ERROR: Please edit/recreate the specfile (%s) and set a different output name (e.g. "dist").' %
            CONF['spec']
        )
    return True


def _make_clean_directory(path):
    """
    Create a clean directory from the given directory name.
    """
    if _check_path_overlap(path):
        if os.path.isdir(path) or os.path.isfile(path):
            try:
                os.remove(path)
            except OSError:
                _rmtree(path)

        os.makedirs(path, exist_ok=True)


def _rmtree(path):
    """
    Remove directory and all its contents, but only after user confirmation, or if the -y option is set.
    """
    from PyInstaller.config import CONF
    if CONF['noconfirm']:
        choice = 'y'
    elif sys.stdout.isatty():
        choice = input(
            'WARNING: The output directory "%s" and ALL ITS CONTENTS will be REMOVED! Continue? (y/N)' % path
        )
    else:
        raise SystemExit(
            'ERROR: The output directory "%s" is not empty. Please remove all its contents or use the -y option (remove'
            ' output directory without confirmation).' % path
        )
    if choice.strip().lower() == 'y':
        if not CONF['noconfirm']:
            print("On your own risk, you can use the option `--noconfirm` to get rid of this question.")
        logger.info('Removing dir %s', path)
        shutil.rmtree(path)
    else:
        raise SystemExit('User aborted')


# TODO Refactor to prohibit empty target directories. As the docstring below documents, this function currently permits
# the second item of each 2-tuple in "hook.datas" to be the empty string, in which case the target directory defaults to
# the source directory's basename. However, this functionality is very fragile and hence bad. Instead:
#
# * An exception should be raised if such item is empty.
# * All hooks currently passing the empty string for such item (e.g.,
#   "hooks/hook-babel.py", "hooks/hook-matplotlib.py") should be refactored
#   to instead pass such basename.
def format_binaries_and_datas(binaries_or_datas, workingdir=None):
    """
    Convert the passed list of hook-style 2-tuples into a returned set of `TOC`-style 2-tuples.

    Elements of the passed list are 2-tuples `(source_dir_or_glob, target_dir)`.
    Elements of the returned set are 2-tuples `(target_file, source_file)`.
    For backwards compatibility, the order of elements in the former tuples are the reverse of the order of elements in
    the latter tuples!

    Parameters
    ----------
    binaries_or_datas : list
        List of hook-style 2-tuples (e.g., the top-level `binaries` and `datas` attributes defined by hooks) whose:
        * The first element is either:
          * A glob matching only the absolute or relative paths of source non-Python data files.
          * The absolute or relative path of a source directory containing only source non-Python data files.
        * The second element is the relative path of the target directory into which these source files will be
          recursively copied.

        If the optional `workingdir` parameter is passed, source paths may be either absolute or relative; else, source
        paths _must_ be absolute.
    workingdir : str
        Optional absolute path of the directory to which all relative source paths in the `binaries_or_datas`
        parameter will be prepended by (and hence converted into absolute paths) _or_ `None` if these paths are to be
        preserved as relative. Defaults to `None`.

    Returns
    ----------
    set
        Set of `TOC`-style 2-tuples whose:
        * First element is the absolute or relative path of a target file.
        * Second element is the absolute or relative path of the corresponding source file to be copied to this target
          file.
    """
    toc_datas = set()

    for src_root_path_or_glob, trg_root_dir in binaries_or_datas:
        # Disallow empty source path. Those are typically result of errors, and result in implicit collection of the
        # whole current working directory, which is never a good idea.
        if not src_root_path_or_glob:
            raise InvalidSrcDestTupleError(
                (src_root_path_or_glob, trg_root_dir),
                "Empty SRC is not allowed when adding binary and data files, as it would result in collection of the "
                "whole current working directory."
            )
        if not trg_root_dir:
            raise InvalidSrcDestTupleError(
                (src_root_path_or_glob, trg_root_dir),
                "Empty DEST_DIR is not allowed - to collect files into application's top-level directory, use "
                f"{os.curdir!r}."
            )
        # Disallow absolute target paths, as well as target paths that would end up pointing outside of the
        # application's top-level directory.
        if os.path.isabs(trg_root_dir):
            raise InvalidSrcDestTupleError((src_root_path_or_glob, trg_root_dir), "DEST_DIR must be a relative path!")
        if os.path.normpath(trg_root_dir).startswith('..'):
            raise InvalidSrcDestTupleError(
                (src_root_path_or_glob, trg_root_dir),
                "DEST_DIR must not point outside of application's top-level directory!",
            )

        # Convert relative to absolute paths if required.
        if workingdir and not os.path.isabs(src_root_path_or_glob):
            src_root_path_or_glob = os.path.join(workingdir, src_root_path_or_glob)

        # Normalize paths.
        src_root_path_or_glob = os.path.normpath(src_root_path_or_glob)

        # If given source path is a file or directory path, pass it on.
        # If not, treat it as a glob and pass on all matching paths. However, we need to preserve the directories
        # captured by the glob - as opposed to collecting their contents into top-level target directory. Therefore,
        # we set a flag which is used in subsequent processing to distinguish between original directory paths and
        # directory paths that were captured by the glob.
        if os.path.isfile(src_root_path_or_glob) or os.path.isdir(src_root_path_or_glob):
            src_root_paths = [src_root_path_or_glob]
            was_glob = False
        else:
            src_root_paths = glob.glob(src_root_path_or_glob)
            was_glob = True

        if not src_root_paths:
            raise SystemExit(f'ERROR: Unable to find {src_root_path_or_glob!r} when adding binary and data files.')

        for src_root_path in src_root_paths:
            if os.path.isfile(src_root_path):
                # Normalizing the result to remove redundant relative paths (e.g., removing "./" from "trg/./file").
                toc_datas.add((
                    os.path.normpath(os.path.join(trg_root_dir, os.path.basename(src_root_path))),
                    os.path.normpath(src_root_path),
                ))
            elif os.path.isdir(src_root_path):
                for src_dir, src_subdir_basenames, src_file_basenames in os.walk(src_root_path):
                    # Ensure the current source directory is a subdirectory of the passed top-level source directory.
                    # Since os.walk() does *NOT* follow symlinks by default, this should be the case. (But let's make
                    # sure.)
                    assert src_dir.startswith(src_root_path)

                    # Relative path of the current target directory, obtained by:
                    #
                    # * Stripping the top-level source directory from the current source directory (e.g., removing
                    #   "/top" from "/top/dir").
                    # * Normalizing the result to remove redundant relative paths (e.g., removing "./" from
                    #   "trg/./file").
                    if was_glob:
                        # Preserve directories captured by glob.
                        rel_dir = os.path.relpath(src_dir, os.path.dirname(src_root_path))
                    else:
                        rel_dir = os.path.relpath(src_dir, src_root_path)
                    trg_dir = os.path.normpath(os.path.join(trg_root_dir, rel_dir))

                    for src_file_basename in src_file_basenames:
                        src_file = os.path.join(src_dir, src_file_basename)
                        if os.path.isfile(src_file):
                            # Normalize the result to remove redundant relative paths (e.g., removing "./" from
                            # "trg/./file").
                            toc_datas.add((
                                os.path.normpath(os.path.join(trg_dir, src_file_basename)), os.path.normpath(src_file)
                            ))

    return toc_datas


def get_code_object(modname, filename, optimize):
    """
    Get the code-object for a module.

    This is a simplifed non-performant version which circumvents __pycache__.
    """

    # Once upon a time, we compiled dummy code objects for PEP-420 namespace packages. We do not do that anymore.
    assert filename not in {'-', None}, "Called with PEP-420 namespace package!"

    _, ext = os.path.splitext(filename)
    ext = ext.lower()

    if ext == '.pyc':
        # The module is available in binary-only form. Read the contents of .pyc file using helper function, which
        # supports reading from either stand-alone or archive-embedded .pyc files.
        logger.debug('Reading code object from .pyc file %s', filename)
        pyc_data = _read_pyc_data(filename)
        code_object = marshal.loads(pyc_data[16:])
    else:
        # Assume this is a source .py file, but allow an arbitrary extension (other than .pyc, which is taken in
        # the above branch). This allows entry-point scripts to have an arbitrary (or no) extension, as tested by
        # the `test_arbitrary_ext` in `test_basic.py`.
        logger.debug('Compiling python script/module file %s', filename)

        with open(filename, 'rb') as f:
            source = f.read()

        # If entry-point script has no suffix, append .py when compiling the source. In POSIX builds, the executable
        # has no suffix either; this causes issues with `traceback` module, as it tries to read the executable file
        # when trying to look up the code for the entry-point script (when current working directory contains the
        # executable).
        _, ext = os.path.splitext(filename)
        if not ext:
            logger.debug("Appending .py to compiled entry-point name...")
            filename += '.py'

        try:
            code_object = compile(source, filename, 'exec', optimize=optimize)
        except SyntaxError:
            logger.warning("Sytnax error while compiling %s", filename)
            raise

    return code_object


def replace_filename_in_code_object(code_object, filename):
    """
    Recursively replace the `co_filename` in the given code object and code objects stored in its `co_consts` entries.
    Primarily used to anonymize collected code objects, i.e., by removing the build environment's paths from them.
    """

    consts = tuple(
        replace_filename_in_code_object(const_co, filename) if isinstance(const_co, types.CodeType) else const_co
        for const_co in code_object.co_consts
    )

    return code_object.replace(co_consts=consts, co_filename=filename)


def _should_include_system_binary(binary_tuple, exceptions):
    """
    Return True if the given binary_tuple describes a system binary that should be included.

    Exclude all system library binaries other than those with "lib-dynload" in the destination or "python" in the
    source, except for those matching the patterns in the exceptions list. Intended to be used from the Analysis
    exclude_system_libraries method.
    """
    dest = binary_tuple[0]
    if dest.startswith(f'python{sys.version_info.major}.{sys.version_info.minor}/lib-dynload'):
        return True
    src = binary_tuple[1]
    if fnmatch.fnmatch(src, '*python*'):
        return True
    if not src.startswith('/lib') and not src.startswith('/usr/lib'):
        return True
    for exception in exceptions:
        if fnmatch.fnmatch(dest, exception):
            return True
    return False


def compile_pymodule(name, src_path, workpath, optimize, code_cache=None):
    """
    Given the name and source file for a pure-python module, compile the module in the specified working directory,
    and return the name of resulting .pyc file. The paths in the resulting .pyc module are anonymized by having their
    absolute prefix removed.

    If a .pyc file with matching name already exists in the target working directory, it is re-used (provided it has
    compatible bytecode magic in the header, and that its modification time is newer than that of the source file).

    If the specified module is available in binary-only form, the input .pyc file is copied to the target working
    directory and post-processed. If the specified module is available in source form, it is compiled only if
    corresponding code object is not available in the optional code-object cache; otherwise, it is copied from cache
    and post-processed. When compiling the module, the specified byte-code optimization level is used.

    It is up to caller to ensure that the optional code-object cache contains only code-objects of target optimization
    level, and that if the specified working directory already contains .pyc files, that they were created with target
    optimization level.
    """

    # Construct the target .pyc filename in the workpath
    split_name = name.split(".")
    if "__init__" in src_path:
        # __init__ module; use "__init__" as module name, and construct parent path using all components of the
        # fully-qualified name
        parent_dirs = split_name
        mod_basename = "__init__"
    else:
        # Regular module; use last component of the fully-qualified name as module name, and the rest as the parent
        # path.
        parent_dirs = split_name[:-1]
        mod_basename = split_name[-1]
    pyc_path = os.path.join(workpath, *parent_dirs, mod_basename + '.pyc')

    # Check if optional cache contains module entry
    code_object = code_cache.get(name, None) if code_cache else None

    if code_object is None:
        _, ext = os.path.splitext(src_path)
        ext = ext.lower()

        if ext == '.py':
            # Source py file; read source and compile it.
            with open(src_path, 'rb') as f:
                src_data = f.read()
            code_object = compile(src_data, src_path, 'exec', optimize=optimize)
        elif ext == '.pyc':
            # The module is available in binary-only form. Read the contents of .pyc file using helper function, which
            # supports reading from either stand-alone or archive-embedded .pyc files.
            pyc_data = _read_pyc_data(src_path)
            # Unmarshal code object; this is necessary if we want to strip paths from it
            code_object = marshal.loads(pyc_data[16:])
        else:
            raise ValueError(f"Invalid python module file {src_path}; unhandled extension {ext}!")

    # Replace co_filename in code object with anonymized filename that does not contain full path. Construct the
    # relative filename from module name, similar how we earlier constructed the `pyc_path`.
    co_filename = os.path.join(*parent_dirs, mod_basename + '.py')
    code_object = replace_filename_in_code_object(code_object, co_filename)

    # Write complete .pyc module to in-memory stream. Then, check if .pyc file already exists, compare contents, and
    # (re)write it only if different. This avoids unnecessary (re)writing of the file, and in turn also avoids
    # unnecessary cache invalidation for targets that make use of the .pyc file (e.g., PKG, COLLECT).
    with io.BytesIO() as pyc_stream:
        pyc_stream.write(compat.BYTECODE_MAGIC)
        pyc_stream.write(struct.pack('<I', 0b01))  # PEP-552: hash-based pyc, check_source=False
        pyc_stream.write(b'\00' * 8)  # Zero the source hash
        marshal.dump(code_object, pyc_stream)
        pyc_data = pyc_stream.getvalue()

    if os.path.isfile(pyc_path):
        with open(pyc_path, 'rb') as fh:
            existing_pyc_data = fh.read()
        if pyc_data == existing_pyc_data:
            return pyc_path  # Return path to (existing) file.

    # Ensure the existence of parent directories for the target pyc path
    os.makedirs(os.path.dirname(pyc_path), exist_ok=True)

    # Write
    with open(pyc_path, 'wb') as fh:
        fh.write(pyc_data)

    # Return output path
    return pyc_path


def _read_pyc_data(filename):
    """
    Helper for reading data from .pyc files. Supports both stand-alone and archive-embedded .pyc files. Used by
    `compile_pymodule` and `get_code_object` helper functions.
    """
    src_file = pathlib.Path(filename)

    if src_file.is_file():
        # Stand-alone .pyc file.
        pyc_data = src_file.read_bytes()
    else:
        # Check if .pyc file is stored in a .zip archive, as is the case for stdlib modules in embeddable
        # python on Windows.
        parent_zip_file = misc.path_to_parent_archive(src_file)
        if parent_zip_file is not None and zipfile.is_zipfile(parent_zip_file):
            with zipfile.ZipFile(parent_zip_file, 'r') as zip_archive:
                # NOTE: zip entry names must be in POSIX format, even on Windows!
                zip_entry_name = str(src_file.relative_to(parent_zip_file).as_posix())
                pyc_data = zip_archive.read(zip_entry_name)
        else:
            raise FileNotFoundError(f"Cannot find .pyc file {filename!r}!")

        # Verify the python version
        if pyc_data[:4] != compat.BYTECODE_MAGIC:
            raise ValueError(f"The .pyc module {filename} was compiled for incompatible version of python!")

    return pyc_data


def postprocess_binaries_toc_pywin32(binaries):
    """
    Process the given `binaries` TOC list to apply work around for `pywin32` package, fixing the target directory
    for collected extensions.
    """
    # Ensure that all files collected from `win32`  or `pythonwin` into top-level directory are put back into
    # their corresponding directories. They end up in top-level directory because `pywin32.pth` adds both
    # directories to the `sys.path`, so they end up visible as top-level directories. But these extensions
    # might in fact be linked against each other, so we should preserve the directory layout for consistency
    # between modulegraph-discovered extensions and linked binaries discovered by link-time dependency analysis.
    # Within the same framework, also consider `pywin32_system32`, just in case.
    PYWIN32_SUBDIRS = {'win32', 'pythonwin', 'pywin32_system32'}

    processed_binaries = []
    for dest_name, src_name, typecode in binaries:
        dest_path = pathlib.PurePath(dest_name)
        src_path = pathlib.PurePath(src_name)

        if dest_path.parent == pathlib.PurePath('.') and src_path.parent.name.lower() in PYWIN32_SUBDIRS:
            dest_path = pathlib.PurePath(src_path.parent.name) / dest_path
            dest_name = str(dest_path)

        processed_binaries.append((dest_name, src_name, typecode))

    return processed_binaries


def postprocess_binaries_toc_pywin32_anaconda(binaries):
    """
    Process the given `binaries` TOC list to apply work around for Anaconda `pywin32` package, fixing the location
    of collected `pywintypes3X.dll` and `pythoncom3X.dll`.
    """
    # The Anaconda-provided `pywin32` package installs three copies of `pywintypes3X.dll` and `pythoncom3X.dll`,
    # located in the following directories (relative to the environment):
    # - Library/bin
    # - Lib/site-packages/pywin32_system32
    # - Lib/site-packages/win32
    #
    # This turns our dependency scanner and directory layout preservation mechanism into a lottery based on what
    # `pywin32` modules are imported and in what order. To keep things simple, we deal with this insanity by
    # post-processing the `binaries` list, modifying the destination of offending copies, and let the final TOC
    # list normalization deal with potential duplicates.
    DLL_CANDIDATES = {
        f"pywintypes{sys.version_info[0]}{sys.version_info[1]}.dll",
        f"pythoncom{sys.version_info[0]}{sys.version_info[1]}.dll",
    }

    DUPLICATE_DIRS = {
        pathlib.PurePath('.'),
        pathlib.PurePath('win32'),
    }

    processed_binaries = []
    for dest_name, src_name, typecode in binaries:
        # Check if we need to divert - based on the destination base name and destination parent directory.
        dest_path = pathlib.PurePath(dest_name)
        if dest_path.name.lower() in DLL_CANDIDATES and dest_path.parent in DUPLICATE_DIRS:
            dest_path = pathlib.PurePath("pywin32_system32") / dest_path.name
            dest_name = str(dest_path)

        processed_binaries.append((dest_name, src_name, typecode))

    return processed_binaries


def create_base_library_zip(filename, modules_toc, code_cache=None):
    """
    Create a zip archive with python modules that are needed during python interpreter initialization.
    """
    with zipfile.ZipFile(filename, 'w') as zf:
        for name, src_path, typecode in modules_toc:
            # Obtain code object from cache, or compile it.
            code = None if code_cache is None else code_cache.get(name, None)
            if code is None:
                optim_level = {'PYMODULE': 0, 'PYMODULE-1': 1, 'PYMODULE-2': 2}[typecode]
                code = get_code_object(name, src_path, optimize=optim_level)
            # Determine destination name
            dest_name = name.replace('.', os.sep)
            # Special case: packages have an implied `__init__` filename that needs to be added.
            basename, ext = os.path.splitext(os.path.basename(src_path))
            if basename == '__init__':
                dest_name += os.sep + '__init__'
            dest_name += '.pyc'  # Always .pyc, regardless of optimization level.
            # Replace full-path co_filename in code object with `dest_name` (and shorten suffix from .pyc to .py).
            code = replace_filename_in_code_object(code, dest_name[:-1])
            # Write the .pyc module
            with io.BytesIO() as fc:
                fc.write(compat.BYTECODE_MAGIC)
                fc.write(struct.pack('<I', 0b01))  # PEP-552: hash-based pyc, check_source=False
                fc.write(b'\00' * 8)  # Match behavior of `building.utils.compile_pymodule`
                marshal.dump(code, fc)
                # Use a ZipInfo to set timestamp for deterministic build.
                info = zipfile.ZipInfo(dest_name)
                zf.writestr(info, fc.getvalue())



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Test bootloader behavior for threading code. The default behavior of Python interpreter is to wait for all threads
# before exiting the main process. Bootloader should behave in the same way.

import os
import sys
import threading

_OUT_EXPECTED = ['ONE', 'TWO', 'THREE']

# Code for the subprocess.
if 'PYI_THREAD_TEST_CASE' in os.environ:

    class TestThreadClass(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)

        def run(self):
            print('ONE')
            print('TWO')
            print('THREE')

    # Main process should not exit before the thread stops. This is the behaviour of Python interpreter.
    TestThreadClass().start()

# Execute itself in a subprocess.
else:
    # Differentiate subprocess code.
    itself = sys.argv[0]
    # Run subprocess.
    import subprocess

    # Preserve environment to avoid `Failed to initialize Windows random API (CryptoGen)`
    env = dict(os.environ)
    env['PYI_THREAD_TEST_CASE'] = 'yes'

    proc = subprocess.Popen([itself], stdout=subprocess.PIPE, env=env, stderr=subprocess.PIPE, shell=False)
    # Waits for subprocess to complete.
    out, err = proc.communicate()

    # Make output from subprocess visible.
    print(out)
    out = out.decode('ascii')
    print(out)

    # Remove empty lines from output.
    out = out.strip().splitlines()
    for line in out:
        if not line.strip():  # Empty line detected.
            out.remove(line)
    # Check output.
    if out != _OUT_EXPECTED:
        print(" +++++++ SUBPROCESS ERROR OUTPUT +++++++")
        print(err)
        raise SystemExit(
            'Subprocess did not print ONE, TWO, THREE in correct order. (output was %r, return code was %s)' %
            (out, proc.returncode)
        )



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

name = 'pyi_testmod_relimp.B.C'

# Import pyi_testmod_relimp.B.D
from . import D  # noqa: E402
# Import pyi_testmod_relimp.B.D.X
from .D import X  # noqa: E402, F401
# Import pyi_testmod_relimp.E
from .. import E  # noqa: E402
# Import pyi_testmod_relimp.F.G
from ..F import G  # noqa: E402
# Import pyi_testmod_relimp.F.H
from ..F import H  # noqa: E402

assert D.name == 'pyi_testmod_relimp.B.D'
assert E.name == 'pyi_testmod_relimp.E'
assert G.name == 'pyi_testmod_relimp.F.G'
assert H.name == 'pyi_testmod_relimp.F.H'



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import multiprocessing


class SendEventProcess(multiprocessing.Process):
    def __init__(self, queue):
        multiprocessing.Process.__init__(self)
        self.queue = queue

    def run(self):
        print('SendEventProcess: begin')
        self.queue.put((1, 2))
        print('SendEventProcess: end')


def main(start_method):
    # Set start method
    multiprocessing.set_start_method(start_method)

    # Create a queue, and run a subprocess that fills it with data
    print('Main: begin')
    queue = multiprocessing.Queue()
    process = SendEventProcess(queue)
    process.start()

    results = queue.get()
    print(f'Main: retrieved results: {results}')
    assert results == (1, 2)

    process.join()
    print('Main: end')

    # Ensure process finished successfully
    assert process.exitcode == 0, f"Process exited with non-success code {process.exitcode}!"


if __name__ == '__main__':
    multiprocessing.freeze_support()

    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <start-method>")
        sys.exit(1)

    main(sys.argv[1])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from __future__ import with_statement
try:
    from xml.sax import make_parser
    from xml.sax.handler import ContentHandler
except ImportError:
    has_xml = False
    ContentHandler = object
else:
    has_xml = True
import os, sys, re
from waflib.Tools import cxx
from waflib import Build, Task, Utils, Options, Errors, Context
from waflib.TaskGen import feature, after_method, extension, before_method
from waflib.Configure import conf
from waflib import Logs

MOC_H = ['.h', '.hpp', '.hxx', '.hh']
EXT_RCC = ['.qrc']
EXT_UI = ['.ui']
EXT_QT5 = ['.cpp', '.cc', '.cxx', '.C']


class qxx(Task.classes['cxx']):
    def __init__(self, *k, **kw):
        Task.Task.__init__(self, *k, **kw)
        self.moc_done = 0

    def runnable_status(self):
        if self.moc_done:
            return Task.Task.runnable_status(self)
        else:
            for t in self.run_after:
                if not t.hasrun:
                    return Task.ASK_LATER
            self.add_moc_tasks()
            return Task.Task.runnable_status(self)

    def create_moc_task(self, h_node, m_node):
        try:
            moc_cache = self.generator.bld.moc_cache
        except AttributeError:
            moc_cache = self.generator.bld.moc_cache = {}
        try:
            return moc_cache[h_node]
        except KeyError:
            tsk = moc_cache[h_node] = Task.classes['moc'](env=self.env, generator=self.generator)
            tsk.set_inputs(h_node)
            tsk.set_outputs(m_node)
            tsk.env.append_unique('MOC_FLAGS', '-i')
            if self.generator:
                self.generator.tasks.append(tsk)
            gen = self.generator.bld.producer
            gen.outstanding.append(tsk)
            gen.total += 1
            return tsk
        else:
            delattr(self, 'cache_sig')

    def add_moc_tasks(self):
        node = self.inputs[0]
        bld = self.generator.bld
        if bld.is_install == Build.UNINSTALL:
            return
        try:
            self.signature()
        except KeyError:
            pass
        else:
            delattr(self, 'cache_sig')
        include_nodes = [node.parent] + self.generator.includes_nodes
        moctasks = []
        mocfiles = set()
        for d in bld.raw_deps.get(self.uid(), []):
            if not d.endswith('.moc'):
                continue
            if d in mocfiles:
                continue
            mocfiles.add(d)
            h_node = None
            base2 = d[:-4]
            prefix = node.name[:node.name.rfind('.')]
            if base2 == prefix:
                h_node = node
            else:
                for x in include_nodes:
                    for e in MOC_H:
                        h_node = x.find_node(base2 + e)
                        if h_node:
                            break
                    else:
                        continue
                    break
            if h_node:
                m_node = h_node.change_ext('.moc')
            else:
                raise Errors.WafError('No source found for %r which is a moc file' % d)
            task = self.create_moc_task(h_node, m_node)
            moctasks.append(task)
        self.run_after.update(set(moctasks))
        self.moc_done = 1


class trans_update(Task.Task):
    run_str = '${QT_LUPDATE} ${SRC} -ts ${TGT}'
    color = 'BLUE'


class XMLHandler(ContentHandler):
    def __init__(self):
        ContentHandler.__init__(self)
        self.buf = []
        self.files = []

    def startElement(self, name, attrs):
        if name == 'file':
            self.buf = []

    def endElement(self, name):
        if name == 'file':
            self.files.append(str(''.join(self.buf)))

    def characters(self, cars):
        self.buf.append(cars)


@extension(*EXT_RCC)
def create_rcc_task(self, node):
    rcnode = node.change_ext('_rc.%d.cpp' % self.idx)
    self.create_task('rcc', node, rcnode)
    cpptask = self.create_task('cxx', rcnode, rcnode.change_ext('.o'))
    try:
        self.compiled_tasks.append(cpptask)
    except AttributeError:
        self.compiled_tasks = [cpptask]
    return cpptask


@extension(*EXT_UI)
def create_uic_task(self, node):
    try:
        uic_cache = self.bld.uic_cache
    except AttributeError:
        uic_cache = self.bld.uic_cache = {}
    if node not in uic_cache:
        uictask = uic_cache[node] = self.create_task('ui5', node)
        uictask.outputs = [node.parent.find_or_declare(self.env.ui_PATTERN % node.name[:-3])]


@extension('.ts')
def add_lang(self, node):
    self.lang = self.to_list(getattr(self, 'lang', [])) + [node]


@feature('qt5')
@before_method('process_source')
def process_mocs(self):
    lst = self.to_nodes(getattr(self, 'moc', []))
    self.source = self.to_list(getattr(self, 'source', []))
    for x in lst:
        prefix = x.name[:x.name.rfind('.')]
        moc_target = 'moc_%s.%d.cpp' % (prefix, self.idx)
        moc_node = x.parent.find_or_declare(moc_target)
        self.source.append(moc_node)
        self.create_task('moc', x, moc_node)


@feature('qt5')
@after_method('apply_link')
def apply_qt5(self):
    if getattr(self, 'lang', None):
        qmtasks = []
        for x in self.to_list(self.lang):
            if isinstance(x, str):
                x = self.path.find_resource(x + '.ts')
            qmtasks.append(self.create_task('ts2qm', x, x.change_ext('.%d.qm' % self.idx)))
        if getattr(self, 'update', None) and Options.options.trans_qt5:
            cxxnodes = [a.inputs[0] for a in self.compiled_tasks
                        ] + [a.inputs[0] for a in self.tasks if a.inputs and a.inputs[0].name.endswith('.ui')]
            for x in qmtasks:
                self.create_task('trans_update', cxxnodes, x.inputs)
        if getattr(self, 'langname', None):
            qmnodes = [x.outputs[0] for x in qmtasks]
            rcnode = self.langname
            if isinstance(rcnode, str):
                rcnode = self.path.find_or_declare(rcnode + ('.%d.qrc' % self.idx))
            t = self.create_task('qm2rcc', qmnodes, rcnode)
            k = create_rcc_task(self, t.outputs[0])
            self.link_task.inputs.append(k.outputs[0])
    lst = []
    for flag in self.to_list(self.env.CXXFLAGS):
        if len(flag) < 2:
            continue
        f = flag[0:2]
        if f in ('-D', '-I', '/D', '/I'):
            if (f[0] == '/'):
                lst.append('-' + flag[1:])
            else:
                lst.append(flag)
    self.env.append_value('MOC_FLAGS', lst)


@extension(*EXT_QT5)
def cxx_hook(self, node):
    return self.create_compiled_task('qxx', node)


class rcc(Task.Task):
    color = 'BLUE'
    run_str = '${QT_RCC} -name ${tsk.rcname()} ${SRC[0].abspath()} ${RCC_ST} -o ${TGT}'
    ext_out = ['.h']

    def rcname(self):
        return os.path.splitext(self.inputs[0].name)[0]

    def scan(self):
        if not has_xml:
            Logs.error('No xml.sax support was found, rcc dependencies will be incomplete!')
            return ([], [])
        parser = make_parser()
        curHandler = XMLHandler()
        parser.setContentHandler(curHandler)
        with open(self.inputs[0].abspath(), 'r') as f:
            parser.parse(f)
        nodes = []
        names = []
        root = self.inputs[0].parent
        for x in curHandler.files:
            nd = root.find_resource(x)
            if nd:
                nodes.append(nd)
            else:
                names.append(x)
        return (nodes, names)

    def quote_flag(self, x):
        return x


class moc(Task.Task):
    color = 'BLUE'
    run_str = '${QT_MOC} ${MOC_FLAGS} ${MOCCPPPATH_ST:INCPATHS} ${MOCDEFINES_ST:DEFINES} ${SRC} ${MOC_ST} ${TGT}'

    def quote_flag(self, x):
        return x


class ui5(Task.Task):
    color = 'BLUE'
    run_str = '${QT_UIC} ${SRC} -o ${TGT}'
    ext_out = ['.h']


class ts2qm(Task.Task):
    color = 'BLUE'
    run_str = '${QT_LRELEASE} ${QT_LRELEASE_FLAGS} ${SRC} -qm ${TGT}'


class qm2rcc(Task.Task):
    color = 'BLUE'
    after = 'ts2qm'

    def run(self):
        txt = '\n'.join(['<file>%s</file>' % k.path_from(self.outputs[0].parent) for k in self.inputs])
        code = '<!DOCTYPE RCC><RCC version="1.0">\n<qresource>\n%s\n</qresource>\n</RCC>' % txt
        self.outputs[0].write(code)


def configure(self):
    self.find_qt5_binaries()
    self.set_qt5_libs_dir()
    self.set_qt5_libs_to_check()
    self.set_qt5_defines()
    self.find_qt5_libraries()
    self.add_qt5_rpath()
    self.simplify_qt5_libs()
    if not has_xml:
        Logs.error('No xml.sax support was found, rcc dependencies will be incomplete!')
    if 'COMPILER_CXX' not in self.env:
        self.fatal('No CXX compiler defined: did you forget to configure compiler_cxx first?')
    frag = '#include <QMap>\nint main(int argc, char **argv) {QMap<int,int> m;return m.keys().size();}\n'
    uses = 'QT5CORE'
    for flag in [[], '-fPIE', '-fPIC', '-std=c++11', ['-std=c++11', '-fPIE'], ['-std=c++11', '-fPIC']]:
        msg = 'See if Qt files compile '
        if flag:
            msg += 'with %s' % flag
        try:
            self.check(features='qt5 cxx', use=uses, uselib_store='qt5', cxxflags=flag, fragment=frag, msg=msg)
        except self.errors.ConfigurationError:
            pass
        else:
            break
    else:
        self.fatal('Could not build a simple Qt application')
    if Utils.unversioned_sys_platform() == 'freebsd':
        frag = '#include <QMap>\nint main(int argc, char **argv) {QMap<int,int> m;return m.keys().size();}\n'
        try:
            self.check(
                features='qt5 cxx cxxprogram',
                use=uses,
                fragment=frag,
                msg='Can we link Qt programs on FreeBSD directly?'
            )
        except self.errors.ConfigurationError:
            self.check(
                features='qt5 cxx cxxprogram',
                use=uses,
                uselib_store='qt5',
                libpath='/usr/local/lib',
                fragment=frag,
                msg='Is /usr/local/lib required?'
            )


@conf
def find_qt5_binaries(self):
    env = self.env
    opt = Options.options
    qtdir = getattr(opt, 'qtdir', '')
    qtbin = getattr(opt, 'qtbin', '')
    paths = []
    if qtdir:
        qtbin = os.path.join(qtdir, 'bin')
    if not qtdir:
        qtdir = self.environ.get('QT5_ROOT', '')
        qtbin = self.environ.get('QT5_BIN') or os.path.join(qtdir, 'bin')
    if qtbin:
        paths = [qtbin]
    if not qtdir:
        paths = self.environ.get('PATH', '').split(os.pathsep)
        paths.extend(['/usr/share/qt5/bin', '/usr/local/lib/qt5/bin'])
        try:
            lst = Utils.listdir('/usr/local/Trolltech/')
        except OSError:
            pass
        else:
            if lst:
                lst.sort()
                lst.reverse()
                qtdir = '/usr/local/Trolltech/%s/' % lst[0]
                qtbin = os.path.join(qtdir, 'bin')
                paths.append(qtbin)
    cand = None
    prev_ver = ['5', '0', '0']
    for qmk in ('qmake-qt5', 'qmake5', 'qmake'):
        try:
            qmake = self.find_program(qmk, path_list=paths)
        except self.errors.ConfigurationError:
            pass
        else:
            try:
                version = self.cmd_and_log(qmake + ['-query', 'QT_VERSION']).strip()
            except self.errors.WafError:
                pass
            else:
                if version:
                    new_ver = version.split('.')
                    if new_ver > prev_ver:
                        cand = qmake
                        prev_ver = new_ver
    if not cand:
        try:
            self.find_program('qtchooser')
        except self.errors.ConfigurationError:
            pass
        else:
            cmd = self.env.QTCHOOSER + ['-qt=5', '-run-tool=qmake']
            try:
                version = self.cmd_and_log(cmd + ['-query', 'QT_VERSION'])
            except self.errors.WafError:
                pass
            else:
                cand = cmd
    if cand:
        self.env.QMAKE = cand
    else:
        self.fatal('Could not find qmake for qt5')
    self.env.QT_HOST_BINS = qtbin = self.cmd_and_log(self.env.QMAKE + ['-query', 'QT_HOST_BINS']).strip()
    paths.insert(0, qtbin)

    def find_bin(lst, var):
        if var in env:
            return
        for f in lst:
            try:
                ret = self.find_program(f, path_list=paths)
            except self.errors.ConfigurationError:
                pass
            else:
                env[var] = ret
                break

    find_bin(['uic-qt5', 'uic'], 'QT_UIC')
    if not env.QT_UIC:
        self.fatal('cannot find the uic compiler for qt5')
    self.start_msg('Checking for uic version')
    uicver = self.cmd_and_log(env.QT_UIC + ['-version'], output=Context.BOTH)
    uicver = ''.join(uicver).strip()
    uicver = uicver.replace('Qt User Interface Compiler ', '').replace('User Interface Compiler for Qt', '')
    self.end_msg(uicver)
    if uicver.find(' 3.') != -1 or uicver.find(' 4.') != -1:
        self.fatal('this uic compiler is for qt3 or qt4, add uic for qt5 to your path')
    find_bin(['moc-qt5', 'moc'], 'QT_MOC')
    find_bin(['rcc-qt5', 'rcc'], 'QT_RCC')
    find_bin(['lrelease-qt5', 'lrelease'], 'QT_LRELEASE')
    find_bin(['lupdate-qt5', 'lupdate'], 'QT_LUPDATE')
    env.UIC_ST = '%s -o %s'
    env.MOC_ST = '-o'
    env.ui_PATTERN = 'ui_%s.h'
    env.QT_LRELEASE_FLAGS = ['-silent']
    env.MOCCPPPATH_ST = '-I%s'
    env.MOCDEFINES_ST = '-D%s'


@conf
def set_qt5_libs_dir(self):
    env = self.env
    qtlibs = getattr(Options.options, 'qtlibs', None) or self.environ.get('QT5_LIBDIR')
    if not qtlibs:
        try:
            qtlibs = self.cmd_and_log(env.QMAKE + ['-query', 'QT_INSTALL_LIBS']).strip()
        except Errors.WafError:
            qtdir = self.cmd_and_log(env.QMAKE + ['-query', 'QT_INSTALL_PREFIX']).strip()
            qtlibs = os.path.join(qtdir, 'lib')
    self.msg('Found the Qt5 libraries in', qtlibs)
    env.QTLIBS = qtlibs


@conf
def find_single_qt5_lib(self, name, uselib, qtlibs, qtincludes, force_static):
    env = self.env
    if force_static:
        exts = ('.a', '.lib')
        prefix = 'STLIB'
    else:
        exts = ('.so', '.lib')
        prefix = 'LIB'

    def lib_names():
        for x in exts:
            for k in ('', '5') if Utils.is_win32 else ['']:
                for p in ('lib', ''):
                    yield (p, name, k, x)

    for tup in lib_names():
        k = ''.join(tup)
        path = os.path.join(qtlibs, k)
        if os.path.exists(path):
            if env.DEST_OS == 'win32':
                libval = ''.join(tup[:-1])
            else:
                libval = name
            env.append_unique(prefix + '_' + uselib, libval)
            env.append_unique('%sPATH_%s' % (prefix, uselib), qtlibs)
            env.append_unique('INCLUDES_' + uselib, qtincludes)
            env.append_unique('INCLUDES_' + uselib, os.path.join(qtincludes, name.replace('Qt5', 'Qt')))
            return k
    return False


@conf
def find_qt5_libraries(self):
    env = self.env
    qtincludes = self.environ.get('QT5_INCLUDES') or self.cmd_and_log(env.QMAKE +
                                                                      ['-query', 'QT_INSTALL_HEADERS']).strip()
    force_static = self.environ.get('QT5_FORCE_STATIC')
    try:
        if self.environ.get('QT5_XCOMPILE'):
            self.fatal('QT5_XCOMPILE Disables pkg-config detection')
        self.check_cfg(atleast_pkgconfig_version='0.1')
    except self.errors.ConfigurationError:
        for i in self.qt5_vars:
            uselib = i.upper()
            if Utils.unversioned_sys_platform() == 'darwin':
                fwk = i.replace('Qt5', 'Qt')
                frameworkName = fwk + '.framework'
                qtDynamicLib = os.path.join(env.QTLIBS, frameworkName, fwk)
                if os.path.exists(qtDynamicLib):
                    env.append_unique('FRAMEWORK_' + uselib, fwk)
                    env.append_unique('FRAMEWORKPATH_' + uselib, env.QTLIBS)
                    self.msg('Checking for %s' % i, qtDynamicLib, 'GREEN')
                else:
                    self.msg('Checking for %s' % i, False, 'YELLOW')
                env.append_unique('INCLUDES_' + uselib, os.path.join(env.QTLIBS, frameworkName, 'Headers'))
            else:
                ret = self.find_single_qt5_lib(i, uselib, env.QTLIBS, qtincludes, force_static)
                if not force_static and not ret:
                    ret = self.find_single_qt5_lib(i, uselib, env.QTLIBS, qtincludes, True)
                self.msg('Checking for %s' % i, ret, 'GREEN' if ret else 'YELLOW')
    else:
        path = '%s:%s:%s/pkgconfig:/usr/lib/qt5/lib/pkgconfig:/opt/qt5/lib/pkgconfig:/usr/lib/qt5/lib:/opt/qt5/lib' % (
            self.environ.get('PKG_CONFIG_PATH', ''), env.QTLIBS, env.QTLIBS
        )
        for i in self.qt5_vars:
            self.check_cfg(
                package=i, args='--cflags --libs', mandatory=False, force_static=force_static, pkg_config_path=path
            )


@conf
def simplify_qt5_libs(self):
    env = self.env

    def process_lib(vars_, coreval):
        for d in vars_:
            var = d.upper()
            if var == 'QTCORE':
                continue
            value = env['LIBPATH_' + var]
            if value:
                core = env[coreval]
                accu = []
                for lib in value:
                    if lib in core:
                        continue
                    accu.append(lib)
                env['LIBPATH_' + var] = accu

    process_lib(self.qt5_vars, 'LIBPATH_QTCORE')


@conf
def add_qt5_rpath(self):
    env = self.env
    if getattr(Options.options, 'want_rpath', False):

        def process_rpath(vars_, coreval):
            for d in vars_:
                var = d.upper()
                value = env['LIBPATH_' + var]
                if value:
                    core = env[coreval]
                    accu = []
                    for lib in value:
                        if var != 'QTCORE':
                            if lib in core:
                                continue
                        accu.append('-Wl,--rpath=' + lib)
                    env['RPATH_' + var] = accu

        process_rpath(self.qt5_vars, 'LIBPATH_QTCORE')


@conf
def set_qt5_libs_to_check(self):
    self.qt5_vars = Utils.to_list(getattr(self, 'qt5_vars', []))
    if not self.qt5_vars:
        dirlst = Utils.listdir(self.env.QTLIBS)
        pat = self.env.cxxshlib_PATTERN
        if Utils.is_win32:
            pat = pat.replace('.dll', '.lib')
        if self.environ.get('QT5_FORCE_STATIC'):
            pat = self.env.cxxstlib_PATTERN
        if Utils.unversioned_sys_platform() == 'darwin':
            pat = r"%s\.framework"
        re_qt = re.compile(pat % 'Qt5?(?P<name>.*)' + '$')
        for x in dirlst:
            m = re_qt.match(x)
            if m:
                self.qt5_vars.append("Qt5%s" % m.group('name'))
        if not self.qt5_vars:
            self.fatal('cannot find any Qt5 library (%r)' % self.env.QTLIBS)
    qtextralibs = getattr(Options.options, 'qtextralibs', None)
    if qtextralibs:
        self.qt5_vars.extend(qtextralibs.split(','))


@conf
def set_qt5_defines(self):
    if sys.platform != 'win32':
        return
    for x in self.qt5_vars:
        y = x.replace('Qt5', 'Qt')[2:].upper()
        self.env.append_unique('DEFINES_%s' % x.upper(), 'QT_%s_LIB' % y)


def options(opt):
    opt.add_option(
        '--want-rpath', action='store_true', default=False, dest='want_rpath', help='enable the rpath for qt libraries'
    )
    for i in 'qtdir qtbin qtlibs'.split():
        opt.add_option('--' + i, type='string', default='', dest=i)
    opt.add_option(
        '--translate', action='store_true', help='collect translation strings', dest='trans_qt5', default=False
    )
    opt.add_option(
        '--qtextralibs',
        type='string',
        default='',
        dest='qtextralibs',
        help='additional qt libraries on the system to add to default ones, comma separated'
    )



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Errors, Logs, Task
from waflib.Tools.ccroot import link_task, stlink_task
from waflib.TaskGen import extension
from waflib.Tools import c_preproc

re_lines = re.compile(
    '^[ \t]*(?:%)[ \t]*(ifdef|ifndef|if|else|elif|endif|include|import|define|undef)[ \t]*(.*)\r*$',
    re.IGNORECASE | re.MULTILINE
)


class asm_parser(c_preproc.c_parser):
    def filter_comments(self, node):
        code = node.read()
        code = c_preproc.re_nl.sub('', code)
        code = c_preproc.re_cpp.sub(c_preproc.repl, code)
        return re_lines.findall(code)


class asm(Task.Task):
    color = 'BLUE'
    run_str = '${AS} ${ASFLAGS} ${ASMPATH_ST:INCPATHS} ${DEFINES_ST:DEFINES} ${AS_SRC_F}${SRC} ${AS_TGT_F}${TGT}'

    def scan(self):
        if self.env.ASM_NAME == 'gas':
            return c_preproc.scan(self)
            Logs.warn('There is no dependency scanner for Nasm!')
            return [[], []]
        elif self.env.ASM_NAME == 'nasm':
            Logs.warn('The Nasm dependency scanner is incomplete!')
        try:
            incn = self.generator.includes_nodes
        except AttributeError:
            raise Errors.WafError('%r is missing the "asm" feature' % self.generator)
        if c_preproc.go_absolute:
            nodepaths = incn
        else:
            nodepaths = [x for x in incn if x.is_child_of(x.ctx.srcnode) or x.is_child_of(x.ctx.bldnode)]
        tmp = asm_parser(nodepaths)
        tmp.start(self.inputs[0], self.env)
        return (tmp.nodes, tmp.names)


@extension('.s', '.S', '.asm', '.ASM', '.spp', '.SPP')
def asm_hook(self, node):
    return self.create_compiled_task('asm', node)


class asmprogram(link_task):
    run_str = '${ASLINK} ${ASLINKFLAGS} ${ASLNK_TGT_F}${TGT} ${ASLNK_SRC_F}${SRC}'
    ext_out = ['.bin']
    inst_to = '${BINDIR}'


class asmshlib(asmprogram):
    inst_to = '${LIBDIR}'


class asmstlib(stlink_task):
    pass


def configure(conf):
    conf.env.ASMPATH_ST = '-I%s'



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import pathlib
import shutil

import pytest

from PyInstaller.compat import is_win, is_darwin

# Bring all common fixtures into this file.
from PyInstaller.utils.conftest import *  # noqa: F401, F403


# A fixture that compiles a test shared library from `data/load_dll_using_ctypes/ctypes_dylib.c` in a sub-directory
# of the tmp_dir, and returns path to the compiled shared library.
#
# NOTE: for this fixture to be session-scoped, we need to define it here (as opposed to `PyInstaller.utils.confest`).
# This is because its data directory needs to be resolved based on this module's location. (And even if we were to use
# test-scoped fixture and infer location from `request.path`, the fixture would be valid only for test files from this
# directory).
@pytest.fixture(scope="session")
def compiled_dylib(tmp_path_factory):
    # Copy the source to temporary directory.
    orig_source_dir = pathlib.Path(__file__).parent / 'data' / 'ctypes_dylib'
    tmp_source_dir = tmp_path_factory.mktemp('compiled_ctypes_dylib')
    shutil.copy2(orig_source_dir / 'ctypes_dylib.c', tmp_source_dir)

    # Compile shared library using `distuils.ccompiler` module. The code is loosely based on implementation of the
    # `distutils.command.build_ext` command module.
    def _compile_dylib(source_dir):
        # Until python 3.12, `distutils` was part of standard library. For newer python versions, `setuptools` provides
        # its vendored copy. If neither are available, skip the test.
        try:
            import distutils.ccompiler
            import distutils.sysconfig
        except ImportError:
            pytest.skip('distutils.ccompiler is not available')

        # Set up compiler
        compiler = distutils.ccompiler.new_compiler()
        distutils.sysconfig.customize_compiler(compiler)
        if hasattr(compiler, 'initialize'):  # Applicable to MSVCCompiler on Windows.
            compiler.initialize()

        if is_win:
            # With MinGW compiler, the `customize_compiler()` call ends up changing `compiler.shared_lib_extension` into
            # ".pyd". Use ".dll" instead.
            suffix = '.dll'
        elif is_darwin:
            # On macOS, `compiler.shared_lib_extension` is ".so", but ".dylib" is more appropriate.
            suffix = '.dylib'
        else:
            suffix = compiler.shared_lib_extension

        # Change the current working directory to the directory that contains source files. Ideally, we could pass the
        # absolute path to sources to `compiler.compile()` and set its `output_dir` argument to the directory where
        # object files should be generated. However, in this case, the object files are put under output directory
        # *while retaining their original path component*. If `output_dir` is not specified, then the original absolute
        # source file paths seem to be turned into relative ones (e.g, on Linux, the leading / is stripped away).
        #
        # NOTE: with python >= 3.11 we could use contextlib.chdir().
        old_cwd = pathlib.Path.cwd()
        os.chdir(source_dir)
        try:
            # Compile source .c file into object
            sources = [
                'ctypes_dylib.c',
            ]
            objects = compiler.compile(sources)

            # Link into shared library
            output_filename = f'ctypes_dylib{suffix}'
            compiler.link_shared_object(
                objects,
                output_filename,
                target_lang='c',
                export_symbols=['add_twelve'],
            )
        finally:
            os.chdir(old_cwd)  # Restore old working directory.

        # Return path to compiled shared library
        return source_dir / output_filename

    try:
        return _compile_dylib(tmp_source_dir)
    except Exception as e:
        pytest.skip(f"Could not compile test shared library: {e}")



# ----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# ----------------------------------------------------------------------------

import os
import sys
import subprocess

import pytest

from PyInstaller.compat import is_win, is_cygwin

if is_win:
    # On Windows, only spawn method is supported.
    START_METHODS = ['spawn']
elif is_cygwin:
    # Under Cygwin, forkserver does not seem to work even in unfrozen python.
    START_METHODS = ['spawn', 'fork']
else:
    # On POSIX systems (including macOS), all methods are supported.
    START_METHODS = ['spawn', 'fork', 'forkserver']


@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_process(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_process.py", app_args=[start_method])


@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_queue(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_queue.py", app_args=[start_method])


@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_pool(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_pool.py", app_args=[start_method])


@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_semaphore(pyi_builder, start_method, capfd):
    pyi_builder.test_script("pyi_multiprocessing_semaphore.py", app_args=[start_method])

    out, err = capfd.readouterr()

    # Print the captured output and error so that it will show up in the test output.
    sys.stderr.write(err)
    sys.stdout.write(out)

    expected = ["In main", "In subprocess"]

    assert os.linesep.join(expected) in out
    for substring in expected:
        assert out.count(substring) == 1


# Test that we can start `multiprocessing.Process` in thread-safe manner concurrently, from multiple threads at (almost)
# the same time. See #7410.
@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_process_start_in_threads(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_process_start_in_threads.py", app_args=[start_method])


# Test that we can start a nested `multiprocessing.Process` from within a `multiprocessing.Process`. See #7494.
# Nested multi-processing is broken in Python 3.11.5 because SemLock.is_fork_ctx attribute (added in
# https://github.com/python/cpython/commit/34ef75d3ef559288900fad008f05b29155eb8b59) is not properly
# serialized/deserialized.
@pytest.mark.parametrize("start_method", START_METHODS)
@pytest.mark.xfail(sys.version_info[:3] == (3, 11, 5), reason="Python 3.11.5 broke nested multiprocessing.")
def test_multiprocessing_nested_process(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_nested_process.py", app_args=[start_method])


# Test that we are able to retrieve the code object for `__main__` module in the sub-process.
# NOTE: in unfrozen version, this works only with `fork` start method. However, in current `multiprocessing` support,
# it should work with all start methods when frozen.
@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_main_module_code_in_process(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_main_module_code_in_process.py", app_args=[start_method])


# Test the basic usage of high-level `concurrent.futures` framework with its `ProcessPoolExecutor` (i.e., with default
# `multiprocessing` start method). This test will be more interesting if/when we can remove the explicit
# `multiprocessing.freeze_support` call in the entry-point script.
def test_concurrent_features_process_pool_executor(pyi_builder):
    pyi_builder.test_source(
        """
        import multiprocessing
        import concurrent.futures

        def square(x):
            return x * x


        if __name__ == '__main__':
            multiprocessing.freeze_support()

            values = range(10)
            with concurrent.futures.ProcessPoolExecutor() as executor:
                result = list(executor.map(square, values))

            assert result == [x * x for x in values]
        """
    )


# Test that application's top level directory (sys._MEIPASS) is properly inherited by multiprocessing child process.
@pytest.mark.parametrize("start_method", START_METHODS)
def test_multiprocessing_subprocess_environment(pyi_builder, start_method):
    pyi_builder.test_script("pyi_multiprocessing_subprocess_environment.py", app_args=[start_method])


# Test the inheritance of application's top level directory (sys._MEIPASS) into sub-processes that are manually spawned
# using `subprocess` module. If using the same executable (`sys.executable`), sys._MEIPASS should be inherited by the
# child process (in onefile mode, this means no unpacking). If it is a different executable, sys._MEIPASS should not be
# inherited (and a onefile child process should unpack itself).
def test_subprocess_environment_inheritance(pyi_builder_spec, tmp_path):
    # Build the spec. This will build a pair of identical onedir programs and a pair of identical onefile programs,
    # which we can then use to test all pertinent combinations. The `pyi_builder_spec` fixture attempts to run the built
    # executable, and for that part, we need to supply the executable name. Since no parameters are passed to the
    # executable, this is essentially no-op and serves just as a sanity check.
    pyi_builder_spec.test_spec('pyi_subprocess_environment_inheritance.spec', app_name='onedir_program_1')

    print("------- Running custom test. -------", file=sys.stderr)

    # "Manually" detertmine the executable paths, as `pyi_builder_spec._find_executables` cannot cope with custom names
    # that are used in the .spec file.
    dist_dir = tmp_path / 'dist'
    exe_suffix = ".exe" if is_win else ""

    onedir_program_1 = dist_dir / "onedir_program_1" / f"onedir_program_1{exe_suffix}"
    onedir_program_2 = dist_dir / "onedir_program_2" / f"onedir_program_2{exe_suffix}"
    onefile_program_1 = dist_dir / f"onefile_program_1{exe_suffix}"
    onefile_program_2 = dist_dir / f"onefile_program_2{exe_suffix}"

    assert onedir_program_1.is_file()
    assert onedir_program_2.is_file()
    assert onefile_program_1.is_file()
    assert onefile_program_1.is_file()

    # Test all relevant combinations; the programs in pairs are functionally identical, so we need to test only one
    # combination (for example, onedir_program_2 exists only so that onedir_program_1 can use it as a child, but the
    # two are otherwise identical).

    print("--- Test: onedir program spawns child via sys.executable...", file=sys.stderr)
    subprocess.check_call([onedir_program_1, 'parent', 'sys.executable'])

    print("--- Test: onefile program spawns child via sys.executable...", file=sys.stderr)
    subprocess.check_call([onefile_program_1, 'parent', 'sys.executable'])

    print("--- Test: onedir program spawns the other onedir program...", file=sys.stderr)
    subprocess.check_call([onedir_program_1, 'parent', onedir_program_2])

    print("--- Test: onedir program spawns onefile program...", file=sys.stderr)
    subprocess.check_call([onedir_program_1, 'parent', onefile_program_1])

    print("--- Test: onefile program spawns the other onefile program...", file=sys.stderr)
    subprocess.check_call([onefile_program_1, 'parent', onefile_program_2])

    print("--- Test: onefile program spawns onedir program...", file=sys.stderr)
    subprocess.check_call([onefile_program_1, 'parent', onedir_program_1])

    # Test the scenarios where we explicitly force independent instance of the same application.
    # NOTE: this applies only to onefile mode
    print("--- Test: onefile program spawns independent instance via sys.executable...", file=sys.stderr)
    subprocess.check_call([onefile_program_1, 'parent', 'sys.executable', '--force-independent'])



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import fnmatch
from PyInstaller.utils.hooks.setuptools import setuptools_info

# Use cached data files list from setuptools_info, and extract relevant bits (to avoid having to call another
# `collect_data_files` and import `setuptools` in isolated process).
datas = [(src_name, dest_name) for src_name, dest_name in setuptools_info.vendored_data
         if fnmatch.fnmatch(src_name, "**/setuptools/_vendor/jaraco/text/*")]



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller import compat
from packaging.version import Version

pygobject_version = Version(compat.importlib_metadata.version("pygobject")).release

hiddenimports = ['gi._error', 'gi._option']

# PyGObject 3.50.0 added support for `asyncio`, and attempts to import inside the `_gi` extension.
if pygobject_version >= (3, 50, 0):
    hiddenimports += ['asyncio']

# PyGobject 3.52.0 added `gi._enum`, which needs to be added to hiddenimports due to being imported from the
# `_gi` extension.
if pygobject_version >= (3, 52, 0):
    hiddenimports += ['gi._enum']



from mypackage import _forbidden_enabled

# Parent level 2
from ... import submod1  # noqa: F401

try:
    from ... import forbidden1  # noqa: F401
except ImportError:
    if _forbidden_enabled:
        raise

# Parent level 1
from .. import submod2  # noqa: F401

try:
    from .. import forbidden2  # noqa: F401
except ImportError:
    if _forbidden_enabled:
        raise

# Current level
from . import submod3  # noqa: F401

try:
    from . import forbidden3  # noqa: F401
except ImportError:
    if _forbidden_enabled:
        raise



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) or, at the user's discretion, the MIT License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception OR MIT)
#-----------------------------------------------------------------------------

import os
import logging
import gc

import pytest

from PyInstaller import isolated
from PyInstaller import compat
from PyInstaller.utils.tests import requires


def add_1(x):
    return x + 1


def fail():
    assert 0, "It's broken!"


def test_basic():
    assert isolated.call(add_1, 10) == 11
    assert isolated.call(add_1, x=4) == 5
    assert isolated.call(lambda: print("hello")) is None


def test_exception_handling():
    """
    Test the behaviour which an error is raised in the child process.
    """
    # Raising an error in a subprocess should propagate to a runtime error in the master process.
    with pytest.raises(RuntimeError) as ex_info:
        isolated.call(fail)
    # The exception should include the offending function's name and the child's traceback which itself includes the
    # breaking line of code and the exception type and message (in that order). Unfortunately, there seems to be a bug
    # in pytest's exception rewriting which makes the exception type and message unverifiable when ran under pytest.
    assert ex_info.match(r"""(?s) call to fail\(\) failed .* assert 0, "It's broken!".*""")
    # It should have removed the following line.
    assert "Traceback (most recent call last)" not in str(ex_info.value)


def test_multiple_calls():
    """
    Test running multiple functions in one process.
    """
    with isolated.Python() as child:
        assert child.call(add_1, 1) == 2
        assert child.call(add_1, 2) == 3
        assert child.call(add_1, 3) == 4


def test_builtins_access():
    """
    Ensure that generic builtins are accessible and that imports work.
    """
    def _use_builtins():
        """
        Test builtin functions, classes and constants are available.
        """
        assert sum([1, 2, 3]) == 6
        list(range(10))
        print("hello")
        Ellipsis
        ...
        NotImplemented

    def _use_imports():
        """
        Test that import-ing is possible.
        """
        import string
        string.digits

        import json
        return json.dumps({'a': 1, 'b': 2})

    with isolated.Python() as child:
        child.call(_use_builtins)
        child.call(_use_imports)


def test_context_wrapping():
    """
    Test the `with` mechanisms of IsolatedPython().

    We can land in some pretty horrible deadlocks if we try talking to a subprocess which either doesn't yet exist or
    has died. Make sure neither of these are possible.

    """
    self = isolated.Python()

    # Don't allow calling a function without using the context manager.
    with pytest.raises(RuntimeError, match="isolated.Python .* 'with' clause"):
        assert self.call(add_1, 1) == 2

    # Multiple enters/exits should work (although they're technically pointless, being equivalent to two
    # isolated.Python() instances).
    with self:
        assert self.call(add_1, 5.1) == 6.1
    with self:
        self.call(add_1, 2) == 3

    # The isolator should remain alive and functional after raising an error.
    with self:
        with pytest.raises(RuntimeError):
            self.call(add_1, "I should be a number")
        assert self.call(add_1, 2.5) == 3.5


def add_spam_to_environ():
    """
    Set an environment variable.
    """
    import os
    os.environ["SPAM"] = "More Spam"
    return "Done it!"


def get_spam_environ():
    import os
    return os.environ.get("SPAM")


def test_environment_propagation(monkeypatch):
    """
    Test that environment variables from the parent process are copied by the child but not the other way around.
    """
    # Clear the SPAM variable globally.
    monkeypatch.delenv("SPAM", raising=False)

    with isolated.Python() as child:
        # There should be no SPAM defined in the child.
        assert child.call(get_spam_environ) is None

        # Define SPAM in the child.
        child.call(add_spam_to_environ)
        # This should change the child environment...
        assert child.call(get_spam_environ) == "More Spam"
        # ... but not the parent.
        assert get_spam_environ() is None

    # Define SPAM globally.
    monkeypatch.setenv("SPAM", "More Spam")
    assert get_spam_environ() == "More Spam"
    # SPAM should be forwarded to new children.
    assert isolated.call(get_spam_environ) == "More Spam"


def test_decorator():
    """
    Test decorating a function with @wrap_with_isolation().
    """
    wrapped = isolated.decorate(add_spam_to_environ)
    assert wrapped.__doc__ == add_spam_to_environ.__doc__

    assert wrapped() == "Done it!"
    assert "SPAM" not in os.environ


@requires("psutil")
def test_pipe_leakage():
    """
    There is a finite number of open pipes/file handles/file descriptors allowed per process. Ensure that all
    opened handles eventually get closed to prevent such *leakages* causing crashes in very long processes (such as
    the rest of our test suite).
    """

    from psutil import Process
    parent = Process()

    # On Windows, the very first `subprocess.Popen()` seems to open two additional handles, which seem to remain open
    # for the duration of the python process. Therefore, if a subprocess is spawned before this test is ran (either as
    # a part of a preceding test, as a part of test collection, or even as a side effect of some 3rd party package
    # having been loaded), those two extra handles will already be part of the initially-opened handles. If not, they
    # will show up during this test and skew the results. Therefore, run a dummy subprocess here to normalize the
    # conditions, and ensure that the test can be ran on its own.
    if compat.is_win:
        import subprocess
        import sys
        proc = subprocess.Popen(
            [sys.executable, '--version'],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        proc.wait()
        del proc

    try:
        # Prevent handles from being non-deterministically closed by garbage collect.
        gc.freeze()

        # Get this platform's *count open handles* method.
        open_fds = parent.num_handles if compat.is_win else parent.num_fds
        old = open_fds()

        # Creating an isolated.Python() does nothing.
        child = isolated.Python()
        assert open_fds() == old

        # On POSIX systems, entering the context creates the child process and 4 handles for sending/receiving to/from
        # it. After creating the child process, we close the descriptors that were passed to the child, so the expected
        # total increase in the parent/main process is two file descriptors.
        #
        # On Windows, the overall setup is a bit more complicated. We create both pipes, spawn the child process, and
        # close end-points (handles) of the pipes that were passed to the child. Then, we open file descriptors on top
        # of the remaining two pipe handles, using `msvcrt.open_osfhandle`. These descriptors are then passed to
        # `os.fdopen` to obtain buffered python `file` object.
        #
        # Instead of monitoring file descriptors, we monitor file handles.
        #
        # Under python 3.12 and earlier, four handles are opened when both pipes are created. Additional two handles are
        # opened when the child process is spawned. Closing the two pipe end-points closes two handles. Opening file
        # descriptors on top of pipe handles (`msvcrt.open_osfhandle`) does not open additional handles, but opening
        # python file objects on top of those descriptors via `os.fdopen` does - one for each file object. This adds two
        # additional file handles, bringing us to the total of six.
        #
        # With python 3.13, the behavior has changed. Creating both pipes still opens four handles. However, spawning
        # the child process opens only one handle. Closing the two pipe end-points closes two handles. No additional
        # handles are created when opening file descriptors on top of pipe handles (`msvcrt.open_osfhandle`), nor when
        # opening python file objects on top of those (`os.fdopen`). So in this case, the total is three.
        if compat.is_win:
            EXPECTED_INCREASE_IN_FDS = 3 if compat.is_py313 else 6
        else:
            EXPECTED_INCREASE_IN_FDS = 2

        with child:
            assert open_fds() == old + EXPECTED_INCREASE_IN_FDS
        # Exiting must close them all immediately. No implicit closure by garbage collect.
        assert open_fds() == old

        # Do it again just to be sure that the context manager properly restarts.
        with child:
            assert open_fds() == old + EXPECTED_INCREASE_IN_FDS
        assert open_fds() == old

    finally:
        gc.unfreeze()


def is_isolated():
    """
    This is the recommended way of testing if you currently in an isolated process.
    """
    return globals().get("__isolated__", False)


def test_is_isolated():
    """
    Verify that the is_isolated() check returns true in isolation and false in this parent process.
    """
    assert is_isolated() is False
    assert isolated.call(is_isolated) is True
    assert is_isolated() is False


def test_default_args():
    """
    Verify that default arguments are properly passed to the isolated function call.
    """
    def isolated_function(arg1='default1', arg2='default2', arg3='default3'):
        return arg1, arg2, arg3

    # Sanity check
    assert isolated_function.__defaults__ == ('default1', 'default2', 'default3')
    assert isolated_function.__kwdefaults__ is None

    # Test by keeping the second argument at the default value
    expected = 'override1', 'default2', 'override3'
    with isolated.Python() as child:
        actual = child.call(isolated_function, arg1='override1', arg3='override3')
    assert actual == expected


def test_default_kwargs():
    """
    Verify that default keyword-only arguments are properly passed to the isolated function call.
    """
    def isolated_function(*args, kwarg1='default1', kwarg2='default2', kwarg3='default3'):
        return kwarg1, kwarg2, kwarg3

    # Sanity check
    assert isolated_function.__defaults__ is None
    assert isolated_function.__kwdefaults__ == {'kwarg1': 'default1', 'kwarg2': 'default2', 'kwarg3': 'default3'}

    # Test by keeping the second keyword-only argument at the default value
    expected = 'override1', 'default2', 'override3'
    with isolated.Python() as child:
        actual = child.call(isolated_function, kwarg1='override1', kwarg3='override3')
    assert actual == expected


@pytest.mark.parametrize("strict_mode", [True, False], ids=['strict', 'lax'])
def test_shutdown_timeout_dangling_threads(strict_mode, caplog):
    """
    Simulate the scenario from #7290, where the isolated sub-process ends up spawning a non-daemon thread, which
    prevents the sub-process from shutting down and in turn blocks the parent process on its _child.wait() call.

    This situation might arise when, as part of analysis, PyInstaller performs an isolated import of a module
    that spawns threads as part of executable statements that are ran on the first time the module is imported.
    """
    def isolated_function(*args):
        import threading
        import time

        # Simulate some long-running background task with periodic activity
        def background_task():
            while True:
                time.sleep(1)

        thread = threading.Thread(target=background_task, name="Test non-daemon thread", daemon=False)
        thread.start()

        return args

    expected = ('a', 'b', 'c', 1, 2, 3)

    if strict_mode:
        # In strict mode, we expect an error to be raised
        with pytest.raises(RuntimeError, match="Timed out while waiting for the child process to exit!"):
            with isolated.Python(strict_mode=strict_mode) as child:
                actual = child.call(isolated_function, *expected)
    else:
        # In lax mode, we expect a warning message
        with caplog.at_level(logging.WARNING):
            with isolated.Python(strict_mode=strict_mode) as child:
                actual = child.call(isolated_function, *expected)

            assert "Timed out while waiting for the child process to exit!" in caplog.text

        # The isolated function should finish and return its expected results, regardless of the shutdown timeout.
        assert actual == expected


def test_subprocess_crash():
    def crash(a):
        import os
        os.kill(os.getpid(), 9)

    with pytest.raises(
        isolated._parent.SubprocessDiedError, match=r"died calling crash\(\) with args=\(12,\) and kwargs=\{\}. .* -?9"
    ):
        isolated.call(crash, 12)


# Nested isolated subprocesses are not supported; attempts to use PyInstaller.isolated from within an isolated
# subprocess should end up reusing the already-existing isolated subprocess. This behavior allows various hook utility
# functions to be transparently used in the same isolated subprocess instead of having them potentially spawn their
# own subprocesses (each of those having to import package(s) again).
def test_nested_isolation():
    def isolated_function():
        from PyInstaller import isolated
        import os

        # Get this process ID of this (isolated process)
        pid = os.getpid()

        @isolated.decorate
        def isolated_subfunction():
            import os
            return os.getpid()

        other_pid = isolated_subfunction()
        return pid, other_pid

    # Test the isolated.call invocation
    pid, other_pid = isolated.call(isolated_function)
    assert pid == other_pid, f"Did not reuse the same isolated process: {pid} vs. {other_pid}"

    # Test the isolated.call invocation
    with isolated.Python() as subprocess:
        pid, other_pid = subprocess.call(isolated_function)
    assert pid == other_pid, f"Did not reuse the same isolated process: {pid} vs. {other_pid}"



import os
import sys
from pathlib import Path

# Import the application's "app" package. This does not(!) contain a sub-module app.hook.
import app  # noqa: F401

# Paths from #4141, where the script was called "main"
#   dist/main/app.py
#   p/plugin_1/app.py
# Paths here:
#   dist/pyi_issue_4141/app.py
#   my-plugins/plugin_11/app.py
dist_main_len = len(os.path.join("dist", "main"))

# Create some "plugins" in sub-directory "p".
#
# Each plugin contains a sub-module called app.hook, which gets imported by the plugin's main module.
# If PyInstaller picks up the application's "app" package instead of the plugin's (which is issue #4141),
# importing the app.hook sub-module will fail.
plugins_dir = Path("p").absolute()
plugin_names = [chr(ord("a") + 10 + i) * (dist_main_len + i) for i in range(5, -5, -1)]
print(plugin_names)

for pn in plugin_names:
    print(plugins_dir / pn / "app")
    (plugins_dir / pn / "app").mkdir(parents=True, exist_ok=True)
    (plugins_dir / pn / "__init__.py").write_text("from .app import hook")
    (plugins_dir / pn / "app" / "hook.py").touch()
del pn

sys.path.insert(0, str(plugins_dir))
print("sys.path[0]:", sys.path[0])

for plugin in plugin_names:
    mod = __import__(plugin)
    print(mod)
    # double check
    assert mod.__file__ == str(plugins_dir / plugin / "__init__.py")



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys
import shutil
from pathlib import Path

import pytest

from PyInstaller.building.icon import normalize_icon_type


# Runs through the normalize_icon_type tests that don't need PIL
def test_normalize_icon(monkeypatch, tmp_path):
    workpath = str(tmp_path)

    # Nonexistent image - FileNotFoundError

    icon = "this_is_not_a_file.ico"
    with pytest.raises(FileNotFoundError):
        normalize_icon_type(icon, ("ico",), "ico", workpath)

    # Native image - file path is passed through unchanged

    icon = str(Path(__file__, "../../functional/data/set_icon/pyi_icon.ico").resolve())
    ret = normalize_icon_type(icon, ("ico",), "ico", workpath)
    if ret != icon:
        pytest.fail("icon validation changed path even though the format was correct already", False)

    # Alternative image - after calling monkeypatch.setitem(sys.modules, "PIL", None): Raise the install pillow error

    monkeypatch.setitem(sys.modules, "PIL", None)
    icon = str(Path(__file__, "../../functional/data/splash/image.png").resolve())
    assert os.path.exists(icon)
    with pytest.raises(ValueError):
        normalize_icon_type(icon, ("ico",), "ico", workpath)


# Runs through the normalize_icon_type tests that DO need PIL
def test_normalize_icon_pillow(tmp_path):
    workpath = str(tmp_path)

    pytest.importorskip("PIL", reason="Needs PIL / Pillow for this test")

    # Alternative image - output is a different file with the correct suffix

    icon = str(Path(__file__, "../../functional/data/splash/image.png").resolve())
    ret = normalize_icon_type(icon, ("ico",), "ico", workpath)

    _, ret_filetype = os.path.splitext(ret)
    if ret_filetype != ".ico":
        pytest.fail("icon validation didn't convert to the right format", False)

    # A .ico which is really a mislabelled .png: should be detected and normalised

    for (i, suffix) in enumerate(["ico", "ICO"]):
        png = shutil.copy(icon, str(tmp_path / f"png-in-disguise-{i}.{suffix}"))
        normalised = normalize_icon_type(png, ("exe", "ico"), "ico", workpath)
        assert normalised != png
        assert normalize_icon_type(normalised, ("exe", "ico"), "ico", workpath) == normalised

    # Some random non-image file: Raises an image conversion error

    icon = os.path.join(tmp_path, 'pyi_icon.notanicon')
    with open(icon, "w", encoding="utf-8") as f:
        f.write("this is in fact, not an icon")

    with pytest.raises(ValueError):
        normalize_icon_type(icon, ("ico",), "ico", workpath)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Errors
from waflib.Tools import ccroot, ar
from waflib.Configure import conf


@conf
def find_scc(conf):
    v = conf.env
    cc = conf.find_program('cc', var='CC')
    try:
        conf.cmd_and_log(cc + ['-flags'])
    except Errors.WafError:
        conf.fatal('%r is not a Sun compiler' % cc)
    v.CC_NAME = 'sun'
    conf.get_suncc_version(cc)


@conf
def scc_common_flags(conf):
    v = conf.env
    v.CC_SRC_F = []
    v.CC_TGT_F = ['-c', '-o', '']
    if not v.LINK_CC:
        v.LINK_CC = v.CC
    v.CCLNK_SRC_F = ''
    v.CCLNK_TGT_F = ['-o', '']
    v.CPPPATH_ST = '-I%s'
    v.DEFINES_ST = '-D%s'
    v.LIB_ST = '-l%s'
    v.LIBPATH_ST = '-L%s'
    v.STLIB_ST = '-l%s'
    v.STLIBPATH_ST = '-L%s'
    v.SONAME_ST = '-Wl,-h,%s'
    v.SHLIB_MARKER = '-Bdynamic'
    v.STLIB_MARKER = '-Bstatic'
    v.cprogram_PATTERN = '%s'
    v.CFLAGS_cshlib = ['-xcode=pic32', '-DPIC']
    v.LINKFLAGS_cshlib = ['-G']
    v.cshlib_PATTERN = 'lib%s.so'
    v.LINKFLAGS_cstlib = ['-Bstatic']
    v.cstlib_PATTERN = 'lib%s.a'


def configure(conf):
    conf.find_scc()
    conf.find_ar()
    conf.scc_common_flags()
    conf.cc_load_tools()
    conf.cc_add_flags()
    conf.link_add_flags()
